# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import gi
from gi.repository import Gtk, GObject, Gdk
from plots.i18n import _

class PopoverColorPicker(Gtk.Button):
    __gtype_name__ = "PopoverColorPicker"
    __gsignals__ = {
        "color-activated": (GObject.SIGNAL_RUN_FIRST, None,
                            (Gtk.ColorChooser, Gdk.RGBA))
    }

    def __init__(self):
        super().__init__()

        self.set_tooltip_text(_("Pick a color"))

        self.chooser = Gtk.ColorChooserWidget.new()
        self.chooser.show()
        self.chooser.connect("color-activated", self.on_color_activated)
        click_ctl = Gtk.GestureClick()
        click_ctl.connect("pressed", self.on_button)
        self.chooser.add_controller(click_ctl)

        self.popover = Gtk.Popover()
        self.popover.set_position(Gtk.PositionType.BOTTOM)
        self.popover.set_parent(self)
        self.popover.set_child(self.chooser)
        self.popover.connect("closed", self.on_close)

        self.connect("clicked", self.on_click)

        self.provider = Gtk.CssProvider()
        context = self.get_style_context()
        context.add_provider(self.provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        self.update_color()


    def update_color(self):
        css = f'button {{ background: {self.get_rgba().to_string()}; }}'
        self.provider.load_from_data(css, len(css))

    def on_click(self, button):
        self.chooser.props.show_editor = False
        self.popover.popup()

    def on_close(self, popover):
        self.on_color_activated(self.chooser, self.get_rgba())

    def get_rgba(self):
        return self.chooser.get_rgba()

    def set_rgba(self, color):
        self.chooser.set_rgba(color)
        self.update_color()

    def add_palette(self, orientation, colors_per_line, colors):
        self.chooser.add_palette(orientation, colors_per_line, colors)

    def on_color_activated(self, chooser, color):
        self.emit("color-activated", chooser, color)
        self.set_rgba(color)

    def on_button(self, ctl, n_press, x, y):
        self.on_color_activated(self.chooser, self.get_rgba())



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import re
from itertools import count

import gi
from gi.repository import GLib, Gtk, Gdk, cairo, Pango, PangoCairo, GObject
from plots.utils import saved, Direction, font_metrics, MetricContext, \
    deitalify_string, ints_to_floats
from plots.data import GREEK_REGEXES, FUNCTIONS, BINARY_OPERATORS, GREEK_LETTERS

DEBUG = False

from . import abstractelement

class ElementList(abstractelement.AbstractElement):
    h_spacing = 0

    def __init__(self, elements=None, parent=None):
        super().__init__(parent)
        if isinstance(elements, ElementList):
            self.elements = elements.elements
        else:
            self.elements = elements or []
        for i, e in enumerate(self.elements):
            e.parent = self
            e.index_in_parent = i

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)

    def __repr__(self):
        return "ElementList({!r})".format(self.elements)

    def __add__(self, other):
        return ElementList(self.elements + other.elements)

    def __getitem__(self, key):
        return self.elements[key]

    def children(self):
        return self.elements

    def compute_metrics(self, ctx, metric_ctx):
        self.ascent = self.descent = self.width = 0
        metric_ctx = MetricContext(metric_ctx.cursor)
        metric_ctx.prev = font_metrics(ctx)
        for i, e in enumerate(self.elements):
            e.index_in_parent = i
            e.compute_metrics(ctx, metric_ctx)
            self.ascent = max(self.ascent, e.ascent)
            self.descent = max(self.descent, e.descent)
            self.width += e.width + 2*e.h_spacing
            metric_ctx.prev = e
        if not self.elements:
            self.ascent = font_metrics(ctx).ascent
            self.descent = font_metrics(ctx).descent
            self.width = font_metrics(ctx).width

    def draw_cursor(self, ctx, ascent, descent, cursor, widget_transform):
        if cursor.owner is self and cursor.visible:
            Gdk.cairo_set_source_rgba(ctx, element.Element.color)
            ctx.set_line_width(max(ctx.device_to_user_distance(cursor.WIDTH, cursor.WIDTH)))
            ctx.move_to(0, descent-2)
            ctx.line_to(0, -ascent+2)
            ctx.move_to(0, 0)
            ctx.stroke()
            cursor.position = widget_transform.transform_point(*ctx.user_to_device(0,0))

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        with saved(ctx):
            for i, e in enumerate(self.elements):
                ctx.move_to(0,0)
                if i == cursor.pos:
                    ascent, descent = e.ascent, e.descent
                    if cursor.pos > 0:
                        ascent = max(ascent, self.elements[i-1].ascent)
                        descent = max(descent, self.elements[i-1].descent)
                    self.draw_cursor(ctx, ascent, descent, cursor, widget_transform)
                ctx.move_to(0, 0)
                ctx.translate(e.h_spacing, 0)
                with saved(ctx):
                    e.draw(ctx, cursor, widget_transform)
                ctx.move_to(0,0)
                ctx.translate(e.width + e.h_spacing, 0)
            if cursor.pos == len(self.elements) > 0:
                self.draw_cursor(ctx, self.elements[-1].ascent, self.elements[-1].descent, cursor, widget_transform)
            elif not self.elements:
                self.draw_cursor(ctx, self.ascent, self.descent, cursor, widget_transform)
                ctx.set_source_rgba(0.5, 0.5, 0.5, 0.2)
                ctx.rectangle(0, -self.ascent, self.width, self.ascent + self.descent)
                ctx.fill()

    def backspace(self, cursor, caller=None, direction=Direction.LEFT):
        if self is not cursor.owner:
            cursor.reparent(self, direction.end())
        if direction is Direction.LEFT:
            shift = -1
        elif direction is Direction.RIGHT:
            shift = 0
        if cursor.pos + shift in range(len(self.elements)):
            target = self.elements[cursor.pos + shift]
            child = target.get_next_child(direction)
            if child is not None:
                cursor.reparent(child, direction.end())
                child.backspace(cursor, direction=direction)
            else:
                cursor.pos += shift
                del self.elements[cursor.pos]
        elif self.parent:
            self.parent.dissolve(cursor, self)

    def replace(self, old, new, cursor, cursor_offset=0):
        if old.parent is self:
            if isinstance(new, ElementList):
                self.elements[old.index_in_parent:old.index_in_parent+1] = new.elements
                for i, e in enumerate(new.elements):
                    e.parent = self
                    e.index_in_parent = old.index_in_parent + i
                if cursor_offset is not None:
                    cursor.reparent(self, old.index_in_parent + cursor_offset)
            else:
                self.elements[old.index_in_parent] = new
                new.parent = self

    def update_children(self):
        for i, e in enumerate(self.elements):
            e.parent = self
            e.index_in_parent = i

    def insert(self, element, cursor):
        self.elements.insert(cursor.pos, element)
        cursor.pos += 1
        self.update_children()
        self.convert_specials(cursor)
        if element.cursor_acceptor is not None:
            cursor.reparent(element.cursor_acceptor, -1)

    def insert_elementlist(self, new, cursor, position, cursor_right=True):
        self.elements[position:position] = new.elements
        self.update_children()
        if cursor_right:
            position += len(new)
        cursor.reparent(self, position)

    def greedy_insert(self, cls, cursor):
        eligible = (atom.Atom, radical.Radical)
        left, right = [], []
        if cursor.pos > 0 and cls.greedy_insert_left and \
           (isinstance(self.elements[cursor.pos-1], eligible) \
            or paren.Paren.is_paren(self.elements[cursor.pos-1], left=False)):
            paren_level = 0
            for n, e in enumerate(self.elements[cursor.pos-1::-1]):
                if paren.Paren.is_paren(e, left=True):
                    paren_level -= 1
                    if paren_level <= 0:
                        break
                elif paren.Paren.is_paren(e, left=False):
                    paren_level += 1
                if atom.Atom.part_of_number(e):
                    continue
                if paren_level <= 0:
                    n -= 1
                    break
            if paren_level >= 0:
                n = max(1, n + 1)
                left = self.elements[cursor.pos - n:cursor.pos]
                del self.elements[cursor.pos - n:cursor.pos]
                cursor.pos -= n
        if cursor.pos < len(self.elements) and cls.greedy_insert_right and \
           (isinstance(self.elements[cursor.pos], eligible) \
            or paren.Paren.is_paren(self.elements[cursor.pos], left=True)):
            paren_level = 0
            for n, e in enumerate(self.elements[cursor.pos:]):
                if paren.Paren.is_paren(e, left=False):
                    paren_level -= 1
                    if paren_level <= 0:
                        break
                elif paren.Paren.is_paren(e, left=True):
                    paren_level += 1
                if atom.Atom.part_of_number(e):
                    continue
                if paren_level <= 0:
                    n -= 1
                    break
            if paren_level >= 0:
                n = max(1, n + 1)
                right = self.elements[cursor.pos:cursor.pos + n]
                del self.elements[cursor.pos:cursor.pos + n]
        new = cls.make_greedily(left, right)
        self.insert(new, cursor)
        cursor.reparent(new.get_next_child(Direction.LEFT if left else Direction.RIGHT), 0)

    def atoms_at_cursor(self, cursor):
        l = cursor.pos
        while l - 1 >= 0:
            if isinstance(self.elements[l-1], (atom.Atom, atom.OperatorAtom)):
                l -= 1
            else:
                break
        r = cursor.pos
        while r < len(self.elements):
            if isinstance(self.elements[r], (atom.Atom, atom.OperatorAtom)):
                r += 1
            else:
                break
        return l, r

    @staticmethod
    def atoms_to_string(atoms):
        return "".join(deitalify_string(atom.name) for atom in atoms)

    def convert_specials(self, cursor):
        l, r = self.atoms_at_cursor(cursor)
        atoms = self.elements[l:r]
        names = index.string_to_names(self.atoms_to_string(atoms))

        # find index of first difference - it will be stored in i
        for i, name, atom in zip(count(), names, atoms):
            if name != deitalify_string(atom.name):
                break
        else:
            return

        new_elems = [index.name_to_element(name) for name in names]
        self.elements[l:r] = new_elems
        for j, elem in enumerate(new_elems):
            elem.parent = self
            elem.index_in_parent = l + j
        if new_elems[i].default_list:
            cursor.reparent(new_elems[i].default_list, 0)
        else:
            cursor.reparent(self, new_elems[i].index_in_parent)
            cursor.handle_movement(Direction.RIGHT)

    def to_glsl(self):
        string_stack = [[]]
        body_stack = [[]]
        sums = []
        sum_paren_levels = []
        prev = None
        parens = 0
        for elem in self.elements:
            if prev is not None and \
               not isinstance(prev, atom.BinaryOperatorAtom) and \
               not isinstance(elem, atom.BinaryOperatorAtom) and \
               not isinstance(elem, supersubscript.SuperscriptSubscript) and \
               not atom.Atom.part_of_number(elem) and \
               not paren.Paren.is_paren(elem, left=False) and \
               not paren.Paren.is_paren(prev, left=True) and \
               not isinstance(prev, atom.OperatorAtom) and \
               not isinstance(prev, sum.Sum) and \
               not (isinstance(elem, atom.Atom) and elem.name == "!"):
                string_stack[-1].append("*")
            # subscript log base
            if isinstance(elem, supersubscript.SuperscriptSubscript) \
               and elem.subscript is not None \
               and isinstance(prev, atom.OperatorAtom) \
               and prev.name == "log":
                assert string_stack[-1].pop() == "log"
                b, e = elem.subscript.to_glsl()
                body_stack[-1].append(b)
                string_stack[-1].append(f"log_base({e},")
                parens += 1
                prev = None
                continue
            if isinstance(prev, atom.OperatorAtom) and \
               not paren.Paren.is_paren(elem, left=True):
                string_stack[-1].append("(")
                parens += 1
            elif isinstance(elem, atom.BinaryOperatorAtom) or paren.Paren.is_paren(elem, left=False):
                string_stack[-1].append(")"*parens)
                parens = 0
            # Postscripts (exponents or factorials)
            if isinstance(elem, supersubscript.SuperscriptSubscript) and elem.exponent is not None \
               or isinstance(elem, atom.Atom) and elem.name == "!":
                parens2 = i = 0
                for i, s in reversed(list(enumerate(string_stack[-1]))):
                    if s == ")":
                        parens2 += 1
                    elif s == "(":
                        parens2 -= 1
                    if parens2 == 0 and s in "+-*=" or parens2 < 0:
                        i += 1
                        break
                if isinstance(elem, supersubscript.SuperscriptSubscript):
                    string_stack[-1].insert(i, "mypow(")
                    b, e = elem.exponent.to_glsl()
                    string_stack[-1].append(f", ({e}))")
                    body_stack[-1].append(b)
                else:
                    string_stack[-1].insert(i, "factorial(")
                    string_stack[-1].append(")")
            elif isinstance(elem, sum.Sum):
                string_stack.append([])
                body_stack.append([])
                sums.append(elem)
                sum_paren_levels.append(0)
            elif not isinstance(elem, supersubscript.SuperscriptSubscript):
                if (isinstance(elem, atom.BinaryOperatorAtom) or
                    paren.Paren.is_paren(elem, left=False)):
                    while sums and sum_paren_levels[-1] == 0:
                        sum_body, sum_expr = sums.pop().to_glsl("".join(body_stack.pop()),
                                                                "".join(string_stack.pop()))
                        sum_paren_levels.pop()
                        string_stack[-1].append(sum_expr)
                        body_stack[-1].append(sum_body)
                elem_body, elem_expr = elem.to_glsl()
                body_stack[-1].append(elem_body)
                string_stack[-1].append(elem_expr)
                if sums and paren.Paren.is_paren(elem, left=True):
                    sum_paren_levels[-1] += 1
                elif sums and paren.Paren.is_paren(elem, left=False):
                    sum_paren_levels[-1] -= 1
            prev = elem
        string_stack[-1].append(")"*parens)
        while sums:
            sum_body, sum_expr = sums.pop().to_glsl("".join(body_stack.pop()),
                                                    "".join(string_stack.pop()))
            sum_paren_levels.pop()
            string_stack[-1].append(sum_expr)
            body_stack[-1].append(sum_body)
        return ints_to_floats("".join(body_stack[-1])), \
            ints_to_floats("".join(string_stack[-1]))

    def to_latex(self):
        return "".join(e.to_latex() for e in self.elements)


from . import sum
from . import paren
from . import radical
from . import frac
from . import supersubscript
from . import atom
from . import index
from . import element



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import freetype
import OpenGL.GL as gl
from OpenGL.GL import shaders
from OpenGL.arrays import vbo
import math
import glm
import importlib.resources as resources
from contextlib import contextmanager


# Code based on:
# - rougier/freetype-py
#   https://github.com/rougier/freetype-py/blob/master/examples/opengl.py
# - LearnOpenGL Text Rendering
#   https://learnopengl.com/In-Practice/Text-Rendering
class TextRenderer():
    def __init__(self, fontsize=14, scale_factor=1):
        self.base, self.texid = 0, 0
        self.width, self.height = 0, 0
        self.characters = []
        self.initgl()
        self.fontsize = fontsize * scale_factor
        self.makefont(resources.open_binary('plots.res', 'DejaVuSans.ttf'),
                      self.fontsize)

    def initgl(self):
        vert = resources.read_text("plots.shaders", "text_vert.glsl")
        frag = resources.read_text("plots.shaders", "text_frag.glsl")
        vert = shaders.compileShader(vert, gl.GL_VERTEX_SHADER)
        frag = shaders.compileShader(frag, gl.GL_FRAGMENT_SHADER)
        self.shaderProgram = shaders.compileProgram(vert, frag)
        self.vbo = vbo.VBO(np.array([
            # x y  u  v
            0, -1, 0, 0,
            0,  0, 0, 1,
            1,  0, 1, 1,
            0, -1, 0, 0,
            1,  0, 1, 1,
            1, -1, 1, 0
        ], 'f'), usage="GL_STATIC_DRAW")
        self.vao = gl.glGenVertexArrays(1)

        gl.glBindVertexArray(self.vao)
        self.vbo.bind()
        self.vbo.copy_data()
        gl.glVertexAttribPointer(0, 4, gl.GL_FLOAT, False, 4*self.vbo.data.itemsize, self.vbo)
        gl.glEnableVertexAttribArray(0)
        self.vbo.unbind()
        gl.glBindVertexArray(0)

    def makefont(self, filename, fontsize):
        face = freetype.Face(filename)
        face.set_pixel_sizes(0, fontsize)

        gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
        gl.glActiveTexture(gl.GL_TEXTURE0)

        self.top_bearing = 0
        for c in range(128):
            face.load_char(chr(c), freetype.FT_LOAD_RENDER)
            glyph = face.glyph
            bitmap = glyph.bitmap
            size = bitmap.width, bitmap.rows
            bearing = glyph.bitmap_left, glyph.bitmap_top
            self.top_bearing = max(self.top_bearing, bearing[1])
            advance = glyph.advance.x

            # create glyph texture
            texObj = gl.glGenTextures(1)
            gl.glBindTexture(gl.GL_TEXTURE_2D, texObj)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE)
            gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_R8, *size, 0, gl.GL_RED, gl.GL_UNSIGNED_BYTE, bitmap.buffer)

            self.characters.append((texObj, size, bearing, advance))

        gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

    def uniform(self, name):
        return gl.glGetUniformLocation(self.shaderProgram, name)

    @contextmanager
    def render(self, width, height):
        self.width, self.height = width, height
        gl.glUseProgram(self.shaderProgram)
        proj = glm.ortho(0, self.width, self.height, 0, -1, 1)
        gl.glUniformMatrix4fv(self.uniform("projection"),
                           1, gl.GL_FALSE, glm.value_ptr(proj))
        yield self

    def width_of(self, text, scale=1):
        return sum((self.characters[ord(c)][3] >> 6)*scale for c in text)

    def render_text(self, text, pos, scale=1, dir=(1,0), halign='left',
                    valign='bottom', text_color=(.0, .0, .0), bg_color=(1., 1., 1.)):
        offset = glm.vec3()
        if halign in ('center', 'right'):
            width = self.width_of(text, scale)
            offset.x -= width
            if halign == 'center':
                offset.x /= 2
        if valign in ('center', 'top'):
            offset.y += self.top_bearing
            if valign == 'center':
                offset.y /= 2
        gl.glActiveTexture(gl.GL_TEXTURE0)
        gl.glBindVertexArray(self.vao)
        angle_rad = math.atan2(dir[1], dir[0])
        rotateM = glm.rotate(glm.mat4(1), angle_rad, glm.vec3(0, 0, 1))
        transOriginM = glm.translate(glm.mat4(1), glm.vec3(*pos, 0) + offset)
        gl.glUniform3f(self.uniform("fg_color"), *text_color)
        gl.glUniform3f(self.uniform("bg_color"), *bg_color)
        char_x = 0
        for c in text:
            c = ord(c)
            ch = self.characters[c]
            w, h = ch[1][0] * scale, ch[1][1] * scale
            xrel, yrel = char_x + ch[2][0] * scale, (ch[1][1] - ch[2][1]) * scale
            char_x += (ch[3] >> 6) * scale
            scaleM = glm.scale(glm.mat4(1), glm.vec3(w, h, 1))
            transRelM = glm.translate(glm.mat4(1), glm.vec3(xrel, yrel, 0))
            modelM = transOriginM * rotateM * transRelM * scaleM

            gl.glUniformMatrix4fv(self.uniform("model"),
                               1, gl.GL_FALSE, glm.value_ptr(modelM))
            gl.glBindTexture(gl.GL_TEXTURE_2D, ch[0])
            gl.glDrawArrays(gl.GL_TRIANGLES, 0, 6)

        gl.glBindVertexArray(0)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from lark import Lark, Transformer

from plots import elements
from plots.data import GREEK_LETTERS, SUP_ATOMS, SUB_ATOMS, SUP_TRAFO, SUB_TRAFO

class LatexTransformer(Transformer):

    def list(self, items):
        return elements.ElementList(elements=items)
        
    def greek2(self, items):
        name = items[0][1:]
        return GREEK_LETTERS[name]  # the letter will be wrapped by atom(...)

    def atom(self, items):
        return elements.Atom(items[0])

    def times(self, _):
        return elements.BinaryOperatorAtom("×")

    def plus(self, _):
        return elements.BinaryOperatorAtom("+")

    def minus(self, _):
        return elements.BinaryOperatorAtom("−")

    def equals(self, _):
        return elements.BinaryOperatorAtom("=")

    def operator(self, items):
        return elements.OperatorAtom(items[0].value)

    def superscript(self, items):
        return elements.SuperscriptSubscript(exponent=items[0])

    def subscript(self, items):
        return elements.SuperscriptSubscript(subscript=items[0])

    def subscriptsuperscript(self, items):
        return elements.SuperscriptSubscript(subscript=items[0], exponent=items[1])

    def superscriptsubscript(self, items):
        return elements.SuperscriptSubscript(exponent=items[0], subscript=items[1])

    def supatom(self, items):
        return elements.Atom(items[0].translate(SUP_TRAFO))

    def subatom(self, items):
        return elements.Atom(items[0].translate(SUB_TRAFO))

    def supparen(self, items):
        return elements.Paren(items[0].translate(SUP_TRAFO))

    def subparen(self, items):
        return elements.Paren(items[0].translate(SUB_TRAFO))

    def frac(self, items):
        return elements.Frac(numerator=items[0], denominator=items[1])

    def sqrt(self, items):
        return elements.Radical(items[0])

    def nthroot(self, items):
        return elements.Radical(items[1], index=items[0])

    def abs(self, items):
        return elements.Abs(items[0])

    def floor(self, items):
        return elements.Floor(items[0])

    def ceil(self, items):
        return elements.Ceil(items[0])

    def paren(self, items):
        return elements.Paren(items[0].value.replace("\\", ""))

    def sum(self, items):
        return elements.Sum(char="∑", bottom=items[0], top=items[1])

    def prod(self, items):
        return elements.Sum(char="∏", bottom=items[0], top=items[1])

latex_parser = Lark(fr"""
list : (element | ( element supersubpostfix ))*  //allowing two consecutive supersubpostfix-tokens leads to ambiguous parser-results
?blist : "{{" list "}}"
?element : atom
         | binary
         | operator
         | supersub
         | frac
         | radical
         | abs
         | floor
         | ceil
         | paren
         | sum
         | prod
         | subscriptsuperscript
         | superscriptsubscript

GREEK_LETTER_NAME : "\\" ( {" | ".join(['"'+letter+'"' for letter in GREEK_LETTERS])} )
greek2 : GREEK_LETTER_NAME
GREEK : "α".."ω" | "Α".."Ω"
SYMBOL : "." | "!"
atom : LETTER | GREEK | DIGIT | SYMBOL | greek2

TIMES : "\\times" | "*"
binary : TIMES -> times
       | "+" -> plus
       | "-" -> minus
       | "=" -> equals

OPNAME : LETTER+
operator : "\\operatorname{{" OPNAME "}}"

atomaslist.-1 : atom -> list
?argument : blist | atomaslist

subscriptsuperscript.2 : "_" argument "^" argument
superscriptsubscript.2 : "^" argument "_" argument
supersub : "_" argument -> subscript
         | "^" argument -> superscript

SUP_ATOM : {" | ".join(['"'+atom+'"' for atom in SUP_ATOMS])}
SUB_ATOM : {" | ".join(['"'+atom+'"' for atom in SUB_ATOMS])}
supatom : SUP_ATOM
subatom : SUB_ATOM
supop : "⁺" -> plus
      | "⁻" -> minus
subop : "₊" -> plus
      | "₋" -> minus
SUP_PAREN : "⁽" | "⁾"
SUB_PAREN : "₍" | "₎"
supparen : SUP_PAREN
subparen : SUB_PAREN
?supelement : supatom | supop | supparen
?subelement : subatom | subop | subparen
suplist : supelement+ -> list
sublist : subelement+ -> list
supersubpostfix : suplist -> superscript
              | sublist -> subscript
              | ( sublist suplist ) -> subscriptsuperscript
              | ( suplist sublist ) -> superscriptsubscript


frac.10 : "\\frac" argument argument

radical : "\\sqrt" argument -> sqrt
        | "\\sqrt" "[" list "]" argument -> nthroot

abs : "\\abs" argument
    | "\\left"? "|" list "\\right"? "|"
floor : "\\floor" argument
      | "\\left"? "\\lfloor" list "\\right"? "\\rfloor"
ceil : "\\ceil" argument
     | "\\left"? "\\lceil" list "\\right"? "\\rceil"

PAREN : "(" | "[" | "\\{{" | ")" | "]" | "\\}}"
paren : ("\\left"|"\\right")? PAREN

sum : "\\sum" "_" argument "^" argument
prod : "\\prod" "_" argument "^" argument

%import common.LETTER
%import common.DIGIT

%import common.WS
%ignore WS
""", start='list') # lalr seems to have problems with ']'

transformer = LatexTransformer()
def from_latex(string):
    return transformer.transform(latex_parser.parse(string))



from . import elements
from . import element
from plots.utils import saved, Text

class AbstractWrapped(element.Element):
    """Abstract class for elements consisting of an ElementList
sandwiched between two characters (e.g. Abs, Floor, Ceil).
    """
    def __init__(self, argument, left_bar, right_bar, parent=None):
        super().__init__(parent)
        self.argument = elements.ElementList(argument, self)
        self.left_bar_char = left_bar
        self.right_bar_char = right_bar
        self.lists = [self.argument]
        self.cursor_acceptor = self.argument
        self.class_name = type(self).__name__

    def __repr__(self):
        return f"{self.class_name}({self.argument})"

    def compute_metrics(self, ctx, metric_ctx):
        self.argument.compute_metrics(ctx, metric_ctx)
        self.left_bar = Text(self.left_bar_char, ctx)
        self.right_bar = Text(self.right_bar_char, ctx)
        self.width = self.argument.width + \
            self.left_bar.width + self.right_bar.width
        self.ascent = self.argument.ascent
        self.descent = self.argument.descent
        super().compute_metrics(ctx, metric_ctx)

    def draw_bar(self, ctx, bar):
        bar_height = bar.ink_rect.height
        scale_factor = max(1, (self.ascent + self.descent)/bar_height)
        with saved(ctx):
            ctx.translate(0, -self.ascent)
            ctx.scale(1, scale_factor)
            ctx.translate(0, -bar.ink_rect.y)
            ctx.move_to(0, 0)
            bar.draw(ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        self.draw_bar(ctx, self.left_bar)
        ctx.translate(self.left_bar.width, 0)
        self.argument.draw(ctx, cursor, widget_transform)
        ctx.translate(self.argument.width, 0)
        self.draw_bar(ctx, self.right_bar)

    def accept_selection(self, selection, direction):
        self.argument.elements.extend(selection)
        for x in selection:
            x.parent = self.argument

    def to_glsl(self):
        arg_body, arg_expr = self.argument.to_glsl()
        return arg_body, f"{self.class_name.lower()}({arg_expr})"

    def to_latex(self):
        return rf"\{self.class_name.lower()}{{{self.argument.to_latex()}}}"



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import gi
from gi.repository import Gtk, Gdk, Gio, GdkPixbuf, Adw, GObject

from plots import formula, plots, rowcommands, colorpicker, utils
from plots.data import jinja_env
import re, math
from enum import Enum

# Monkey-patch Gdk.RGBA to fix equality
Gdk.RGBA.__eq__ = Gdk.RGBA.equal

class RowData():
    def id(self):
        return id(self)


class Empty(RowData):
    priority = 0
    def __init__(self, owner, **kwargs):
        self.owner = owner
        self.rgba = None

    def definition(self):
        return ""

    def calculation(self):
        return ""

    @staticmethod
    def accepts(expr):
        return True


class Variable(RowData):
    priority = 50
    def __init__(self, owner, body, expr, rgba=None):
        self.owner = owner
        m = re.match(r'^([a-zA-Z_]\w*) *=(.*)', expr)
        self.name = m.group(1)
        self.body = body
        self.expr = expr

    def definition(self):
        return f"float {self.name} = 0.0;\n"

    def calculation(self):
        return f"{self.body}\n{self.expr};\n"

    @staticmethod
    def accepts(expr):
        m = re.match(r'^([a-zA-Z_]\w*) *=(.*)', expr)
        return m and m.group(1) not in ["x", "y"]


class Slider(RowData):
    priority = 80
    def __init__(self, owner, body, expr, rgba):
        self.owner = owner
        m = re.match(r'^([a-zA-Z_]\w*) *= *([+-]?([0-9]*[.])?[0-9]+)', expr)
        self.name = m.group(1)
        self.value = float(m.group(2))

    def definition(self):
        return f"uniform float {self.name};\n"

    def calculation(self):
        return ""

    @staticmethod
    def accepts(expr):
        m = re.match(r'^([a-zA-Z_]\w*) *= *([+-]?([0-9]*[.])?[0-9]+)', expr)
        return m and m.group(1) not in ["x", "y"]


class Formula(RowData):
    priority = 20
    calculation_template = jinja_env.get_template("formula_calculation.glsl")

    def __init__(self, owner, expr, body, rgba):
        self.owner = owner
        m = re.match(r'^(?:y *=)?(.+)', expr)
        self.expr = m.group(1)
        self.body = body
        self.rgba = rgba

    def definition(self):
        return f"""float formula{self.id()}(float x) {{
    {self.body}
    return {self.expr};
}}"""

    def calculation(self):
        return self.calculation_template.render(formula=self)

    @staticmethod
    def accepts(expr):
        m = re.match(r'^(?:y *=)?(.+)', expr)
        return m and "=" not in m.group(1)


class XFormula(RowData):
    priority = 20
    calculation_template = jinja_env.get_template("x_formula_calculation.glsl")

    def __init__(self, owner, expr, body, rgba):
        self.owner = owner
        m = re.match(r'^x *=(.+)', expr)
        self.expr = m.group(1)
        self.body = body
        self.rgba = rgba

    def definition(self):
        return f"""float formula{self.id()}(float y) {{
    {self.body}
    return {self.expr};
}}"""

    def calculation(self):
        return self.calculation_template.render(formula=self)

    @staticmethod
    def accepts(expr):
        m = re.match(r'^x *=(.+)', expr)
        return bool(m)


class RFormula(RowData):
    priority = 20
    calculation_template = jinja_env.get_template("r_formula_calculation.glsl")

    def __init__(self, owner, expr, body, rgba):
        self.owner = owner
        m = re.match(r'^r *=(.+)', expr)
        self.expr = m.group(1)
        self.body = body
        self.rgba = rgba

    def definition(self):
        return f"""float formula{self.id()}(float theta) {{
    {self.body}
    return {self.expr};
}}"""

    def calculation(self):
        return self.calculation_template.render(formula=self)

    @staticmethod
    def accepts(expr):
        m = re.match(r'^r *=(.+)', expr)
        return bool(m)


class ThetaFormula(RowData):
    priority = 20
    calculation_template = jinja_env.get_template("theta_formula_calculation.glsl")

    def __init__(self, owner, expr, body, rgba):
        self.owner = owner
        m = re.match(r'^theta *=(.+)', expr)
        self.expr = m.group(1)
        self.body = body
        self.rgba = rgba

    def definition(self):
        return f"""float formula{self.id()}(float r) {{
    {self.body}
    return {self.expr};
}}"""

    def calculation(self):
        return self.calculation_template.render(formula=self)

    @staticmethod
    def accepts(expr):
        m = re.match(r'^theta *=(.+)', expr)
        return bool(m)


class ImplicitFormula(RowData):
    priority = 20
    calculation_template = jinja_env.get_template("implicit_formula_calculation.glsl")

    def __init__(self, owner, expr, body, rgba):
        self.owner = owner
        m = re.match(r'^([^=]+)=([^=]+)$', expr)
        self.body = body
        self.rgba = rgba
        self.expr = "({}) - ({})".format(m.group(1), m.group(2));

    @staticmethod
    def accepts(expr):
        m = re.match(r'^([^=]+)=([^=]+)$', expr)
        return bool(m)

    def definition(self):
        return f"""float formula{self.id()}(float x, float y) {{
    {self.body}
    return {self.expr};
}}"""

    def calculation(self):
        return self.calculation_template.render(formula=self)


class RowStatus(Enum):
    GOOD = 1
    BAD = 2
    UNKNOWN = 3


@Gtk.Template(string=utils.read_ui_file("formula_box.ui"))
class FormulaBox(Gtk.Box):
    __gtype_name__ = "FormulaBox"
    __gsignals__ = {
        "dependency_changed": (GObject.SIGNAL_RUN_FIRST, None, ())
    }

    PALETTE      = "blue_5 green_5 yellow_5 orange_5 red_5 purple_5 brown_5 dark_5".split()
    DARK_PALETTE = "blue_1 green_1 yellow_1 orange_1 red_1 purple_1 brown_1 light_2".split()
    _palette_use_next = 0

    delete_button = Gtk.Template.Child()
    viewport = Gtk.Template.Child("editor_viewport")
    button_box = Gtk.Template.Child()
    slider = Gtk.Template.Child()
    slider_upper = Gtk.Template.Child()
    slider_lower = Gtk.Template.Child()
    slider_box = Gtk.Template.Child()

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.data = Empty(self)

        self.style_ctx = self.get_style_context()
        self.PALETTE = [self.style_ctx.lookup_color(c).color for c in self.PALETTE]
        self.DARK_PALETTE = [self.style_ctx.lookup_color(c).color for c in self.DARK_PALETTE]

        self.color_picker = colorpicker.PopoverColorPicker()
        self.button_box.append(self.color_picker)
        self.color_picker.add_palette(Gtk.Orientation.HORIZONTAL, 9, self.PALETTE)
        self.color_picker.set_rgba(self.PALETTE[FormulaBox._palette_use_next])
        FormulaBox._palette_use_next = (FormulaBox._palette_use_next + 1) % len(self.PALETTE)
        self.editor = formula.Editor()
        self.editor.connect("edit", self.edited)
        self.editor.connect("cursor_position", self.cursor_position)
        self.delete_button.connect("clicked", self.delete)
        self.color_picker.connect("color-activated", self.on_color_activated)
        self.slider.connect("value-changed", self.slider_changed)
        self.slider_upper.connect("changed", self.slider_limits_changed)
        self.slider_lower.connect("changed", self.slider_limits_changed)
        self.viewport.set_child(self.editor)
        self.connect("realize", self.on_realize)
        self.editor.grab_focus()
        self.old = self.construct_memory()
        self.use_dark_style = False
        self.row_status = RowStatus.UNKNOWN

    def on_realize(self, widget):
        self.slider_box.hide()

    def delete(self, widget, record=True, replace_if_last=True):
        if record:
            self.app.add_to_history(rowcommands.Delete(self, self.app.rows))
        self.app.rows.remove(self)
        self.app.formula_box.remove(self)
        if not self.app.rows and replace_if_last:
            self.app.add_equation(None, record=False)
        self.emit("dependency_changed")
        self.app.update_shader()

    def cursor_position(self, widget, x, y):
        adj = widget.get_parent().get_hadjustment().props
        # Force adjustment to update to new size
        adj.upper = max(widget.get_size_request()[0], adj.page_size)
        if x - 4 < adj.value:
            adj.value = x - 4
        elif x + 4 > adj.value + adj.page_size:
            adj.value = x - adj.page_size + 4

    def on_color_activated(self, widget, chooser, color):
        self.edited(widget)

    def edited(self, widget, record=True):
        body, expr = self.editor.expr.to_glsl()
        rgba = utils.rgba_to_tuple(self.color_picker.get_rgba())

        old_data = self.data
        for cls in [Formula, XFormula, RFormula, ThetaFormula,
                    Slider, Variable, ImplicitFormula, Empty]:
            if cls.accepts(expr):
                self.data = cls(self, body=body, expr=expr, rgba=rgba)
                break

        if hasattr(self.data, 'rgba'):
            self.color_picker.show()
        else:
            self.color_picker.hide()
            self.name = self.data.name

        if isinstance(self.data, Slider):
            self.slider_box.show()
            val = self.data.value
            if val == 0:
                u, l = 10., -10.
            else:
                u = 10**(1+math.floor(math.log10(abs(val))))
                l = -abs(u)/10
                if val < 0:
                    u, l = -l, -u
                u = max(u, 10.0)
                l = min(l, -10.0)
            self.slider_upper.set_text(str(u))
            self.slider_lower.set_text(str(l))
            self.slider.set_value(val)
        else:
            self.slider_box.hide()

        mem = self.construct_memory()
        if record:
            command = rowcommands.Edit(self, self.app.rows, mem, self.old)
            self.app.add_to_history(command)
        self.old = mem
        self.row_status = RowStatus.UNKNOWN
        if isinstance(self.data, (Slider, Variable)) \
           or isinstance(old_data, (Slider, Variable)):
            self.emit("dependency_changed")
        self.app.update_shader()

    def construct_memory(self):
        adj = self.slider.get_adjustment()
        return rowcommands.RowMemory(
            formula=self.editor.expr.to_latex(),
            rgba=self.color_picker.get_rgba(),
            lower=adj.get_lower(),
            upper=adj.get_upper(),
            slider=adj.get_value())

    def slider_changed(self, widget):
        self.editor.cursor.cancel_selection()
        equals_index = self.editor.expr.elements.index(formula.BinaryOperatorAtom("="))
        del self.editor.expr.elements[equals_index+1:]
        cursor = self.editor.cursor
        cursor.reparent(self.editor.expr, -1)
        new_val = round(self.slider.get_value(), 4)
        for char in str(int(new_val) if new_val.is_integer() else new_val):
            if char == "-":
                self.editor.expr.insert(formula.BinaryOperatorAtom("−"), cursor)
            elif char.isdigit() or char == ".":
                self.editor.expr.insert(formula.Atom(char), cursor)
        self.editor.queue_draw()
        self.app.gl_area.queue_draw()

    def slider_limits_changed(self, widget):
        try:
            if widget is self.slider_upper:
                self.slider.get_adjustment().set_upper(float(widget.get_text()))
            elif widget is self.slider_lower:
                self.slider.get_adjustment().set_lower(float(widget.get_text()))
        except ValueError:
            pass

    @property
    def value(self):
        return self.slider.get_value()

    def get_data(self):
        return self.data

    def do_css_changed(self, change):
        style = self.style_is_dark()
        if style != self.use_dark_style:
            self.use_dark_style = style
            new_palette = [self.PALETTE, self.DARK_PALETTE][self.use_dark_style]
            old_palette = [self.PALETTE, self.DARK_PALETTE][not self.use_dark_style]
            new_color = old_color = self.color_picker.get_rgba()
            if old_color in old_palette:
                new_color = new_palette[old_palette.index(old_color)]
            self.color_picker.add_palette(Gtk.Orientation.HORIZONTAL, 9, None)
            self.color_picker.add_palette(Gtk.Orientation.HORIZONTAL, 9, new_palette)
            self.color_picker.set_rgba(new_color)
            self.edited(None, record=False)

    def style_is_dark(self):
        return Adw.StyleManager.get_default().get_dark()



from . import elements
from . import element
from plots.utils import saved, Direction

class SuperscriptSubscript(element.Element):
    h_spacing = 0
    exponent_scale = 0.7
    subscript_scale = 0.7
    subscript_shift = 6
    superscript_adjustment = 14

    def __init__(self, parent=None, exponent=None, subscript=None):
        super().__init__(parent)
        self.exponent = exponent
        self.subscript = subscript
        self.lists = []
        self._selection_acceptor = None
        self.update_lists()

    def __repr__(self):
        return f"SuperscriptSubscript(exponent={repr(self.exponent)}, subscript={repr(self.subscript)})"

    def add_superscript(self, cursor):
        if self.exponent is None:
            self.exponent = elements.ElementList([], self)
            self.update_lists()
        self.cursor_acceptor = self.exponent
        self._selection_acceptor = self.exponent

    def add_subscript(self, cursor):
        if self.subscript is None:
            self.subscript = elements.ElementList([], self)
            self.update_lists()
        self.cursor_acceptor = self.subscript
        self._selection_acceptor = self.subscript

    def update_lists(self):
        self.lists = [x for x in (self.exponent, self.subscript) if x is not None]
        for l in self.lists:
            l.parent = self

    def compute_metrics(self, ctx, metric_ctx):
        self.width = 0
        self.ascent = max(0, metric_ctx.prev.ascent)
        self.descent = max(0, metric_ctx.prev.descent)
        if self.exponent is not None:
            self.exponent.compute_metrics(ctx, metric_ctx)
            self.superscript_shift = -self.exponent.descent*self.exponent_scale \
                - metric_ctx.prev.ascent + self.superscript_adjustment # -ve y is up
            self.width = max(self.width, self.exponent.width*self.exponent_scale)
            self.ascent = max(self.ascent, self.exponent.ascent*self.exponent_scale - self.superscript_shift)
            self.descent = max(self.descent, self.exponent.descent*self.exponent_scale + self.superscript_shift)
        if self.subscript is not None:
            self.subscript.compute_metrics(ctx, metric_ctx)
            self.width = max(self.width, self.subscript.width*self.subscript_scale)
            self.ascent = max(self.ascent, self.subscript.ascent*self.subscript_scale - self.subscript_shift)
            self.descent = max(self.descent, self.subscript.descent*self.subscript_scale + self.subscript_shift)
        super().compute_metrics(ctx, metric_ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        if self.exponent is not None:
            with saved(ctx):
                ctx.translate(0, self.superscript_shift)
                ctx.scale(self.exponent_scale, self.exponent_scale)
                self.exponent.draw(ctx, cursor, widget_transform)
        if self.subscript is not None:
            with saved(ctx):
                ctx.translate(0, self.subscript_shift)
                ctx.scale(self.subscript_scale, self.subscript_scale)
                self.subscript.draw(ctx, cursor, widget_transform)

    @classmethod
    def make_greedily(cls, left, right):
        return cls(exponent=right)

    def accept_selection(self, selection, direction):
        if direction is Direction.LEFT:
            self._selection_acceptor.elements[0:0] = selection
        else:
            self._selection_acceptor.elements.extend(selection)
        for x in selection:
            x.parent = self._selection_acceptor

    def dissolve(self, cursor, caller):
        if len(self.lists) == 1:
            return super().dissolve(cursor, caller)
        self.lists.remove(caller)
        if caller is self.exponent:
            self.exponent = None
            self.parent.insert_elementlist(caller, cursor, self.index_in_parent+1, False)
        elif caller is self.subscript:
            self.subscript = None
            self.parent.insert_elementlist(caller, cursor, self.index_in_parent, True)

    def to_latex(self):
        res = ""
        if self.subscript is not None:
            res += "_{" + self.subscript.to_latex() + "}"
        if self.exponent is not None:
            res += "^{" + self.exponent.to_latex() + "}"
        return res



from .elements import *
from .abs import Abs
from .sum import Sum
from .paren import Paren
from .radical import Radical
from .frac import Frac
from .supersubscript import SuperscriptSubscript
from .atom import BaseAtom, Atom, BinaryOperatorAtom, OperatorAtom
from .element import Element
from .floor import Floor
from .ceil import Ceil



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from plots import formularow, parser
from collections import namedtuple

RowMemory = namedtuple("RowMemory", "formula rgba lower upper slider")

class Delete():
    def __init__(self, row, rows):
        self.index = rows.index(row)
        self.formula = row.editor.expr.to_latex()
        self.rgba = row.color_picker.get_rgba()
        self.last = len(rows) == 1

    def do(self, app):
        app.rows[self.index].delete(None, record=False)

    def undo(self, app):
        if self.last:
            app.rows[0].delete(None, record=False, replace_if_last=False)
        row = formularow.FormulaBox(app)
        app.insert_row(self.index, row)
        row.editor.set_expr(parser.from_latex(self.formula))
        row.color_picker.set_rgba(self.rgba)
        row.edited(None, record=False)
        row.editor.grab_focus()

class Add():
    def __init__(self, row, rows):
        self.index = rows.index(row)
        self.rgba = row.color_picker.get_rgba()

    def do(self, app):
        row = formularow.FormulaBox(app)
        app.insert_row(self.index, row)
        row.color_picker.set_rgba(self.rgba)
        row.editor.grab_focus()

    def undo(self, app):
        app.rows[-1].delete(None, record=False)

class Edit():
    def __init__(self, row, rows, new, old):
        self.index = rows.index(row)
        self.before = old
        self.after = new

    def do(self, app):
        row = app.rows[self.index]
        row.editor.set_expr(parser.from_latex(self.after.formula))
        row.color_picker.set_rgba(self.after.rgba)
        row.editor.grab_focus()
        row.editor.queue_draw()
        row.edited(None, record=False)

    def undo(self, app):
        row = app.rows[self.index]
        row.editor.set_expr(parser.from_latex(self.before.formula))
        row.color_picker.set_rgba(self.before.rgba)
        row.editor.grab_focus()
        row.editor.queue_draw()
        row.edited(None, record=False)



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import plots.formula as f
from plots.parser import from_latex

lefts = [f.Paren(c) for c in "([{"]
rights = [f.Paren(c) for c in "}])"]

@pytest.mark.parametrize("paren", lefts + rights)
def test_is_paren(paren):
    assert f.Paren.is_paren(paren)

def test_is_not_paren():
    assert not f.Paren.is_paren(f.Atom("f"))

@pytest.mark.parametrize("paren", lefts)
def test_is_left_paren(paren):
    assert f.Paren.is_paren(paren, left=True)

@pytest.mark.parametrize("paren", rights)
def test_is_not_left_paren(paren):
    assert not f.Paren.is_paren(paren, left=True)

@pytest.mark.parametrize("paren", rights)
def test_is_right_paren(paren):
    assert f.Paren.is_paren(paren, left=False)

@pytest.mark.parametrize("paren", lefts)
def test_is_not_right_paren(paren):
    assert not f.Paren.is_paren(paren, left=False)



from plots.utils import Direction

class AbstractElement():
    """Functionality shared by both Element and ElementList
    """
    def __init__(self, parent):
        self.parent = parent

    def contains_device_point(self, x, y):
        return self.top_left[0] <= x <= self.bottom_right[0] and \
            self.top_left[1] <= y <= self.bottom_right[1]

    def half_containing(self, x, y):
        x_mid = (self.bottom_right[0] + self.top_left[0])/2
        if x < x_mid:
            return Direction.LEFT
        else:
            return Direction.RIGHT

    @property
    def height(self):
        return self.ascent + self.descent

    def draw(self, ctx, cursor, widget_transform):
        self.top_left = widget_transform.transform_point(
            *ctx.user_to_device(-self.h_spacing, -self.ascent))
        self.bottom_right = widget_transform.transform_point(
            *ctx.user_to_device(self.width + self.h_spacing, self.descent))



from . import elements
from . import element
from . import atom
from plots.utils import saved, Text

class Sum(element.Element):
    child_scale = 0.7
    bottom_padding = 4
    glsl_var_counter = 0

    def __init__(self, parent=None, char="∑", top=None, bottom=None):
        super().__init__(parent=parent)
        self.top = elements.ElementList(top or [], self)
        self.bottom = elements.ElementList(bottom or [atom.BinaryOperatorAtom("=")], self)
        self.lists = [self.top, self.bottom]
        self.default_list = self.bottom
        self.char = char

    def __repr__(self):
        return f"Sum(char={self.char}, top={self.top}, bottom={self.bottom})"

    def compute_metrics(self, ctx, metric_ctx):
        self.symbol = Text(self.char, ctx, scale=1.5)
        self.top.compute_metrics(ctx, metric_ctx)
        self.bottom.compute_metrics(ctx, metric_ctx)
        self.width = max(self.symbol.width, self.top.width*self.child_scale,
                         self.bottom.width*self.child_scale)
        self.ascent = self.symbol.ascent + self.child_scale*self.top.height
        self.descent = self.symbol.descent + \
            self.child_scale*self.bottom.height + self.bottom_padding
        super().compute_metrics(ctx, metric_ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        with saved(ctx):
            ctx.translate(self.width/2 - self.symbol.width/2, 0)
            self.symbol.draw_at_baseline(ctx)
        with saved(ctx):
            ctx.translate(self.width/2, -self.symbol.ascent)
            ctx.scale(self.child_scale, self.child_scale)
            ctx.translate(-self.top.width/2, -self.top.descent)
            self.top.draw(ctx, cursor, widget_transform)
        with saved(ctx):
            ctx.translate(self.width/2, self.symbol.descent + self.bottom_padding)
            ctx.scale(self.child_scale, self.child_scale)
            ctx.translate(-self.bottom.width/2, self.bottom.ascent)
            self.bottom.draw(ctx, cursor, widget_transform)

    def to_glsl(self, arg_body, arg_expr):
        init_body, init_expr = self.bottom.to_glsl()
        end_body, end_expr = self.top.to_glsl()
        var = init_expr.split('=')[0].strip()
        sum_var = f"sum{Sum.glsl_var_counter}"
        Sum.glsl_var_counter += 1
        if Sum.glsl_var_counter > 100000:
            Sum.glsl_var_counter = 0
        if self.char == "∑":
            init = 0.0
            op = "+="
        elif self.char == "∏":
            init = 1.0
            op = "*="
        body = f"""
        {init_body}
        {end_body}
        float {sum_var} = {init};
        for (float {init_expr}; {var} <= {end_expr}; {var}++) {{
            {arg_body}
            {sum_var} {op} {arg_expr};
        }}"""
        return body, sum_var

    def to_latex(self):
        if self.char == "∑":
            return r"\sum_{" + self.bottom.to_latex() + "}^{" + self.top.to_latex() + "}"
        elif self.char == "∏":
            return r"\prod_{" + self.bottom.to_latex() + "}^{" + self.top.to_latex() + "}"



from . import elements
from . import element
from plots.utils import saved, font_metrics

class Frac(element.Element):
    vertical_separation = 4
    greedy_insert_right = greedy_insert_left = True

    def __init__(self, numerator=None, denominator=None, parent=None):
        super().__init__(parent)
        self.numerator = elements.ElementList(numerator, self)
        self.denominator = elements.ElementList(denominator, self)
        self.lists = [self.numerator, self.denominator]
        self.cursor_acceptor = self.denominator

    def __repr__(self):
        return f"Frac(numerator={self.numerator}, denominator={self.denominator})"

    def compute_metrics(self, ctx, metric_ctx):
        self.numerator.compute_metrics(ctx, metric_ctx)
        self.denominator.compute_metrics(ctx, metric_ctx)
        self.width = max(self.numerator.width, self.denominator.width)

        font_ascent = font_metrics(ctx).ascent
        self.bar_height = font_ascent * 0.3
        self.ascent = self.numerator.ascent + self.numerator.descent + \
            self.bar_height + self.vertical_separation//2
        self.descent = self.denominator.ascent + self.denominator.descent + \
            self.vertical_separation//2 - self.bar_height
        super().compute_metrics(ctx, metric_ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        with saved(ctx):
            ctx.translate(0, -self.bar_height)
            ctx.move_to(0,0)
            ctx.set_line_width(1)
            ctx.line_to(self.width, 0)
            ctx.stroke()
            ctx.move_to(0,0)
            with saved(ctx):
                ctx.translate(self.width//2 - self.numerator.width//2,
                              -self.vertical_separation//2 - self.numerator.descent)
                self.numerator.draw(ctx, cursor, widget_transform)
            with saved(ctx):
                ctx.translate(self.width//2 - self.denominator.width//2,
                              self.vertical_separation//2 + self.denominator.ascent)
                self.denominator.draw(ctx, cursor, widget_transform)

    def accept_selection(self, selection, direction):
        self.numerator.elements.extend(selection)
        for x in selection:
            x.parent = self.numerator

    @classmethod
    def make_greedily(cls, left, right):
        return cls(numerator=left, denominator=right)

    def to_glsl(self):
        num_body, num_expr = self.numerator.to_glsl()
        den_body, den_expr = self.denominator.to_glsl()
        return num_body + den_body, f"({num_expr})/({den_expr})"

    def to_latex(self):
        return "\\frac{" + self.numerator.to_latex() + "}{" + self.denominator.to_latex() + "}"



from . import elements
from . import element
from plots.utils import saved, Text

class Radical(element.Element):
    index_y_shift = 16
    index_x_shift = 16
    index_scale = 0.8

    def __init__(self, radicand, index=None, parent=None):
        super().__init__(parent)
        self.radicand = elements.ElementList(radicand, self)
        if index is not None:
            self.index = elements.ElementList(index, self)
            self.lists = [self.index, self.radicand]
        else:
            self.index = None
            self.lists = [self.radicand]
        self.overline_space = 4

    def __repr__(self):
        return f"Radical({self.radicand}, index={self.index})"

    def compute_metrics(self, ctx, metric_ctx):
        self.radicand.compute_metrics(ctx, metric_ctx)
        self.symbol = Text("√", ctx)
        self.width = self.radicand.width + self.symbol.width
        self.main_ascent = self.ascent = max(self.symbol.ascent, self.radicand.ascent + self.overline_space)
        self.descent = self.radicand.descent
        if self.index is not None:
            self.index.compute_metrics(ctx, metric_ctx)
            self.width += self.index.width*self.index_scale - self.index_x_shift
            self.ascent = max(self.main_ascent,
                              (self.index.ascent + self.index.descent)*self.index_scale \
                              - self.index_y_shift + self.main_ascent)

        super().compute_metrics(ctx, metric_ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)

        if self.index is not None:
            with saved(ctx):
                ctx.translate(0, -self.main_ascent + self.index_y_shift -self.index.descent)
                ctx.scale(self.index_scale, self.index_scale)
                ctx.move_to(0, 0)
                self.index.draw(ctx, cursor, widget_transform)
            ctx.translate(self.index.width*self.index_scale - self.index_x_shift, 0)

        symbol_size = self.symbol.ink_rect.height
        scale_factor = max(1, (self.main_ascent + self.descent)/symbol_size)
        with saved(ctx):
            ctx.translate(0, -self.main_ascent)
            ctx.scale(1, scale_factor)
            ctx.translate(0, -self.symbol.ink_rect.y)
            ctx.move_to(0, 0)
            self.symbol.draw(ctx)

        ctx.translate(self.symbol.width, 0)
        ctx.set_line_width(1)
        ctx.move_to(0, -self.main_ascent + ctx.get_line_width())
        ctx.rel_line_to(self.radicand.width, 0)
        ctx.stroke()
        ctx.move_to(0,0)
        self.radicand.draw(ctx, cursor, widget_transform)

    def to_glsl(self):
        radicand_body, radicand_expr = self.radicand.to_glsl()
        if self.index:
            index_body, index_expr = self.index.to_glsl()
            return radicand_body + index_body, f"pow({radicand_expr}, 1.0/({index_expr}))"
        else:
            return radicand_body, f"sqrt({radicand_expr})"

    def to_latex(self):
        if self.index:
            return "\\sqrt[" + self.index.to_latex() + "]{" + self.radicand.to_latex() + "}"
        else:
            return "\\sqrt{" + self.radicand.to_latex() + "}"



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
import re
import sys

import gi
from gi.repository import GLib, Gtk, Gdk, cairo, Pango, PangoCairo, GObject, Gio
import importlib.resources as resources

desc = Pango.font_description_from_string("Latin Modern Math 20")

class Direction(Enum):
    UP = Gdk.KEY_Up
    DOWN = Gdk.KEY_Down
    LEFT = Gdk.KEY_Left
    RIGHT = Gdk.KEY_Right
    NONE = 0

    def displacement(self):
        if self is self.UP or self is self.LEFT:
            return -1
        elif self is self.DOWN or self is self.RIGHT:
            return 1
        else:
            return 0

    def end(self):
        return -1 if self.displacement() == -1 else 0

    def vertical(self):
        return self is self.UP or self is self.DOWN

    def horizontal(self):
        return not self.vertical()

class saved():
    """Context manager, ensures the cairo context is restored"""
    def __init__(self, ctx):
        self.ctx = ctx

    def __enter__(self):
        self.ctx.save()

    def __exit__(self ,exc_type, exc_val, exc_tb):
        self.ctx.restore()
        return False

class MetricContext():
    "Keeps track of state needed to calculate sizes"
    def __init__(self, cursor=None):
        self.prev = None
        self.paren_stack = []
        self.cursor = cursor


def italify_string(s):
    def italify_char(c):
        if c == 'h':
            return 'ℎ'
        # lowercase latin
        if c.islower() and ord(c) < 128:
            return chr(ord(c) - 0x61 + 0x1d44e)
        # uppercase latin
        if c.isupper() and ord(c) < 128:
            return chr(ord(c) - 0x41 + 0x1d434)
        # lowercase greek (n.b. don't italify uppers)
        if 0x3b1 <= ord(c) < 0x3b1 + 18:
            return chr(ord(c) - 0x3b1 + 0x1d6fc)
        return c
    return "".join(italify_char(c) for c in s)

def deitalify_char(c):
    if c == 'ℎ':
        return 'h'
    if 0x1d44e <= ord(c) < 0x1d44e + 26:
        return chr(ord(c) - 0x1d44e + 0x61)
    if 0x1d434 <= ord(c) < 0x1d434 + 26:
        return chr(ord(c) - 0x1d434 + 0x41)
    if 0x1d6fc <= ord(c) < 0x1d6fc + 18:
        return chr(ord(c) - 0x1d6fc + 0x3b1)
    return c

def deitalify_string(s):
    return "".join(deitalify_char(c) for c in s)


def ints_to_floats(string):
    return re.sub(r"(?<![\.\da-zA-Z_])(\d+)(?![\.\d])", r"\1.0", string)


class Text:
    def __init__(self, text, ctx, scale=1):
        self.scale = scale
        sf = scale/Pango.SCALE
        self.layout = PangoCairo.create_layout(ctx)
        self.layout.set_text(text, -1)
        self.layout.set_font_description(desc)
        self.width, self.height = self.layout.get_size()
        self.width *= sf
        self.height *= sf
        self.ascent = self.layout.get_baseline()*sf
        self.descent = self.height - self.ascent

        # Have to do this because get_pixel_extents returns integer pixels,
        # which are not precise enough
        self.ink_rect, self.logical_rect = self.layout.get_extents()
        for attr in ("x", "y", "width", "height"):
            setattr(self.ink_rect, attr, getattr(self.ink_rect, attr)*sf)
            setattr(self.logical_rect, attr, getattr(self.logical_rect, attr)*sf)

    def draw_at_baseline(self, ctx):
        with saved(ctx):
            ctx.move_to(0, -self.ascent)
            ctx.scale(self.scale, self.scale)
            PangoCairo.show_layout(ctx, self.layout)

    def draw(self, ctx):
        with saved(ctx):
            ctx.scale(self.scale, self.scale)
            self.update()
            PangoCairo.show_layout(ctx, self.layout)

    def update(self):
        self.layout.context_changed()

def font_metrics(ctx):
    return Text("x", ctx)

def rgba_to_tuple(rgba):
    return (rgba.red, rgba.green, rgba.blue, rgba.alpha)

def create_rgba(r, g, b, a=1.0):
    res = Gdk.RGBA()
    res.red = r
    res.green = g
    res.blue = b
    res.alpha = a
    return res

def read_ui_file(name):
    return resources.read_text("plots.ui", name)

def install_excepthook():
    old_hook = sys.excepthook

    def new_hook(etype, evalue, etb):
        old_hook(etype, evalue, etb)
        if (app := Gio.Application.get_default()) is not None:
            app.quit()
        sys.exit()
    sys.excepthook = new_hook

def shader_error_str(error):
    return error.args[0].encode('ascii', 'ignore').decode('unicode_escape')



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import re
from itertools import count

import gi
gi.require_version('PangoCairo', '1.0')
from gi.repository import GLib, Gtk, Gdk, cairo, Pango, PangoCairo, GObject
from enum import Enum
from plots import parser
from plots.elements import Element, ElementList, BaseAtom, Atom, \
    BinaryOperatorAtom, OperatorAtom, SuperscriptSubscript, Frac, Radical, \
    Abs, Paren, Sum
from plots.utils import Direction, MetricContext, Text
from plots.data import BINARY_OPERATORS
import lark.exceptions

class Cursor():
    WIDTH = 1
    BLINK_DELAY = 600

    def __init__(self, editor_widget):
        self.owner = None
        self.editor = editor_widget
        self.visible = True
        self.pos = 0
        self.secondary_pos = None
        self.secondary_owner = None
        self.selecting = False
        self.selection_bounds = None
        self.selection_ancestor = None
        self.selection_rgba = [0.5, 0.5, 1, 0.6]
        self._position = (0., 0.)     # absolute position in widget (in pixels)
        self.position_changed = False  # set to True when self.position changes
        self.clipboard = self.editor.get_clipboard()

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, new):
        if new != self._position:
            self._position = new
            self.position_changed = True

    @property
    def selection_slice(self):
        s = self.selection_bounds
        return slice(s.start, s.stop, s.step)

    def reparent(self, new_parent, position):
        self.owner = new_parent
        self.pos = position
        if position < 0:
            self.pos = len(self.owner.elements) + position + 1

    def cancel_selection(self):
        self.secondary_pos, self.secondary_owner = None, None
        self.selection_bounds, self.selection_ancestor = None, None
        self.selecting = False

    def select_all(self, root):
        self.reparent(root, -1)
        self.secondary_pos = 0
        self.secondary_owner = root
        self.selecting = True
        self.selection_bounds, self.selection_ancestor = self.calculate_selection()

    def copy_selection(self):
        if self.selecting:
            elements = self.selection_ancestor.elements[self.selection_slice]
            text = "".join(e.to_latex() for e in elements)
            self.clipboard.set(text)

    def cut_selection(self):
        self.copy_selection()
        self.backspace(None)

    def paste(self):
        self.clipboard.read_text_async(None, self.paste_cb)

    def paste_cb(self, source, res):
        text = self.clipboard.read_text_finish(res)
        try:
            elements = parser.from_latex(text)
        except lark.exceptions.LarkError:
            return
        if self.selecting:
            self.backspace(None)
        self.owner.insert_elementlist(elements, self, self.pos)
        self.editor.queue_draw()
        self.editor.emit("edit")

    def mouse_select(self, element, direction, drag=False):
        if drag:
            if not self.selecting:
                self.secondary_pos = self.pos
                self.secondary_owner = self.owner
                self.selecting = True
        else:
            self.cancel_selection()
        if not isinstance(element, ElementList):
            pos = element.index_in_parent
            element = element.parent
            if direction is Direction.RIGHT:
                pos += 1
        else:
            if direction is Direction.LEFT:
                pos = 0
            else:
                pos = -1
        self.reparent(element, pos)
        if self.selecting:
            self.selection_bounds, self.selection_ancestor = self.calculate_selection()

    def handle_movement(self, direction, select=False):
        if select and not self.selecting:
            self.secondary_pos, self.secondary_owner = self.pos, self.owner
        elif not select:
            self.cancel_selection()
        self.selecting = select

        shift = 0 if direction.displacement() == 1 else -1

        def go_to_parent():
            if self.owner.parent:
                new_list = self.owner.parent.get_next_child(direction, self.owner)
                if new_list is not None:
                    self.owner = new_list
                    self.pos = len(self.owner) if direction.end() == -1 else 0
                else:
                    self.pos = self.owner.parent.index_in_parent + shift + 1
                    self.owner = self.owner.parent.parent
                return True
            return False

        if direction.vertical():
            return go_to_parent()
        adj_idx = self.pos + shift
        res = True
        try:
            if adj_idx < 0:
                raise IndexError
            adj = self.owner.elements[adj_idx]
            child_list = adj.get_next_child(direction)
            if child_list is not None:
                self.owner = child_list
                self.pos = len(self.owner) if direction.end() == -1 else 0
            else:
                new_pos = self.pos + direction.displacement()
                if new_pos in range(len(self.owner.elements) + 1):
                    self.pos = new_pos
                else:
                    res = False
        except IndexError:
            res = go_to_parent()
        if self.selecting:
            self.selection_bounds, self.selection_ancestor = self.calculate_selection()
        return res

    def calculate_selection(self):
        if self.secondary_owner is None:
            return None
        primary_ancestors = self.ancestors(self.owner)
        secondary_ancestors = self.ancestors(self.secondary_owner)
        for i, l in enumerate(primary_ancestors):
            if l in secondary_ancestors:
                common_ancestor = l
                j = secondary_ancestors.index(l)
                break
        if i > 0:
            primary_ancestor_index = primary_ancestors[i-1].parent.index_in_parent
        else:
            primary_ancestor_index = self.pos
        if j > 0:
            secondary_ancestor_index = secondary_ancestors[j-1].parent.index_in_parent
        else:
            secondary_ancestor_index = self.secondary_pos
        if i > 0 and primary_ancestor_index >= secondary_ancestor_index:
            primary_ancestor_index += 1
        if j > 0 and secondary_ancestor_index >= primary_ancestor_index:
            secondary_ancestor_index += 1
        left = min(primary_ancestor_index, secondary_ancestor_index)
        right = max(primary_ancestor_index, secondary_ancestor_index)
        return range(left, right), common_ancestor

    @staticmethod
    def ancestors(elementlist):
        l = elementlist
        res = [l]
        while l.parent:
            l = l.parent.parent
            res.append(l)
        return res

    def backspace(self, direction):
        if self.selecting:
            sel = self.selection_ancestor.elements[self.selection_bounds.start:self.selection_bounds.stop]
            del self.selection_ancestor.elements[self.selection_bounds.start:self.selection_bounds.stop]
            self.reparent(self.selection_ancestor, self.selection_bounds.start)
            self.cancel_selection()
            return sel
        else:
            return self.owner.backspace(self, direction=direction)

    def insert(self, element, direction=Direction.LEFT):
        self.give_selected(element, direction=direction)
        self.owner.insert(element, self)

    def give_selected(self, element, direction=Direction.LEFT):
        if self.selecting:
            selection = self.backspace(None)
            element.accept_selection(selection, direction=direction)
            return len(selection)

    def greedy_insert(self, cls):
        if self.selecting:
            self.insert(cls())
        else:
            self.owner.greedy_insert(cls, self)

    def insert_superscript_subscript(self, superscript=True):
        new = False
        if self.pos > 0 and isinstance(self.owner.elements[self.pos - 1], SuperscriptSubscript):
            element = self.owner.elements[self.pos - 1]
            direction = Direction.RIGHT
        elif self.pos < len(self.owner) and \
             isinstance(self.owner.elements[self.pos], SuperscriptSubscript):
            element = self.owner.elements[self.pos]
            direction = Direction.LEFT
        elif self.secondary_owner is self.owner and self.secondary_pos > 0 and \
             isinstance(self.owner.elements[self.secondary_pos - 1], SuperscriptSubscript):
            element = self.owner.elements[self.secondary_pos - 1]
            direction = Direction.RIGHT
        elif self.secondary_owner is self.owner and self.secondary_pos < len(self.owner) and \
             isinstance(self.owner.elements[self.secondary_pos], SuperscriptSubscript):
            element = self.owner.elements[self.secondary_pos]
            direction = Direction.LEFT
        else:
            element  = SuperscriptSubscript()
            new = True
            direction = Direction.RIGHT
        if superscript:
            element.add_superscript(self)
        else:
            element.add_subscript(self)
        if new:
            self.insert(element, direction)
        else:
            selection_length = self.give_selected(element, direction)
            if direction is Direction.RIGHT:
                self.reparent(element.cursor_acceptor, -1)
            else:
                self.reparent(element.cursor_acceptor, selection_length or 0)



#!/usr/bin/env python3

# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from gi.repository import Gtk, Adw, GObject

import copy
import os
import configparser

import plots.i18n
from plots import utils

def xdg_config_home():
    varname = "XDG_CONFIG_HOME"
    if varname in os.environ:
        return os.environ[varname]
    else:
        return "{}/.config".format(os.environ["HOME"])


class Preferences(GObject.GObject):
    __gsignals__ = {
        "updated": (GObject.SIGNAL_RUN_FIRST, None, ())
    }
    DEFAULTS = {
        "rendering": {
            "line_thickness": 2.0,
            "samples": 32,
        }
    }
    CONFIG_DIR = "plots"
    CONFIG_FILENAME = "config.ini"

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.data = copy.deepcopy(self.DEFAULTS)
        self.load_config()

    def load_config(self):
        filename = f"{xdg_config_home()}/{self.CONFIG_DIR}/{self.CONFIG_FILENAME}"
        config = configparser.ConfigParser()
        config.read(filename)
        for sec in self.data.keys() & config.keys():
            datasec = self.data[sec]
            for option in datasec.keys() & config[sec].keys():
                if option in datasec:
                    t = type(datasec[option])
                    datasec[option] = t(config[sec][option])

    def save_config(self):
        conf_dir = f"{xdg_config_home()}/{self.CONFIG_DIR}"
        os.makedirs(conf_dir, exist_ok=True)
        filename = f"{conf_dir}/{self.CONFIG_FILENAME}"
        config = configparser.ConfigParser()
        config.read_dict(self.data)
        with open(filename, "w") as f:
            config.write(f)

    def show(self):
        self.window = PreferencesWindow(self, self.parent)
        self.window.connect("close-request", self.close_cb)
        self.window.show()

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def close_cb(self, prefs_window):
        self.emit("updated")
        return False


@Gtk.Template(string=utils.read_ui_file("preferences.ui"))
class PreferencesWindow(Adw.PreferencesWindow):
    __gtype_name__ = "PreferencesWindow"

    line_thickness_scale = Gtk.Template.Child()
    samples_scale = Gtk.Template.Child()

    def __init__(self, prefs, parent_window):
        super().__init__()
        self.prefs = prefs

        self.set_transient_for(parent_window)
        self.props.modal = True
        self.connect("close-request", self.delete_cb)

        self.line_thickness_scale.set_range(1, 10)
        self.line_thickness_scale.set_increments(0.5, 3)
        self.line_thickness_scale.set_value(prefs["rendering"]["line_thickness"])

        self.samples_scale.set_range(4, 128)
        self.samples_scale.set_value(prefs["rendering"]["samples"])
        self.samples_scale.set_digits(0)
        self.samples_scale.set_increments(1, 10)

    def delete_cb(self, window):
        r = self.prefs["rendering"]
        r["line_thickness"] = self.line_thickness_scale.get_value()
        r["samples"] = int(self.samples_scale.get_value())



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from plots.parser import from_latex
from plots.utils import Direction
import plots.formula as f

from tests.fixtures import cursor

def test_insert_empty_super(cursor):
    elems = from_latex("x")
    cursor.reparent(elems, -1)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == r"x^{}"
    assert cursor.owner is elems[1].exponent

def test_add_to_super_after(cursor):
    elems = from_latex("x^{2}2")
    cursor.reparent(elems, -1)
    cursor.handle_movement(Direction.LEFT, select=True)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == "x^{22}"
    assert cursor.owner is elems[1].exponent

def test_add_to_super_after_alternative(cursor):
    elems = from_latex("x^{2}2")
    cursor.reparent(elems, 2)
    cursor.handle_movement(Direction.RIGHT, select=True)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == "x^{22}"
    assert cursor.owner is elems[1].exponent

def test_add_to_super_before(cursor):
    elems = from_latex("xy^{z}")
    cursor.reparent(elems, 1)
    cursor.handle_movement(Direction.RIGHT, select=True)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == "x^{yz}"
    assert cursor.owner is elems[1].exponent

def test_add_to_super_before_alternative(cursor):
    elems = from_latex("xy^{z}")
    cursor.reparent(elems, 2)
    cursor.handle_movement(Direction.LEFT, select=True)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == "x^{yz}"
    assert cursor.owner is elems[1].exponent

def test_move_to_super_before(cursor):
    elems = from_latex("x^{2}")
    cursor.reparent(elems, 1)
    cursor.insert_superscript_subscript()
    assert elems.to_latex() == "x^{2}"
    assert cursor.owner is elems[1].exponent

def test_delete_super(cursor):
    elems = from_latex("x_{4}^{3}")
    cursor.reparent(elems[1].exponent, 0)
    cursor.backspace(Direction.LEFT)
    assert elems.to_latex() == "x_{4}3"

def test_delete_sub(cursor):
    elems = from_latex("x_{4}^{3}")
    cursor.reparent(elems[1].subscript, 0)
    cursor.backspace(Direction.LEFT)
    assert elems.to_latex() == "x4^{3}"

def test_add_sub_to_super(cursor):
    elems = from_latex("x^{3}")
    cursor.reparent(elems, -1)
    cursor.insert_superscript_subscript(superscript=False)
    assert elems.to_latex() == "x_{}^{3}"
    assert cursor.owner is elems[1].subscript

def test_add_super_to_sub(cursor):
    elems = from_latex("x_{3}")
    cursor.reparent(elems, -1)
    cursor.insert_superscript_subscript(superscript=True)
    assert elems.to_latex() == "x_{3}^{}"
    assert cursor.owner is elems[1].exponent



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, GdkPixbuf
import OpenGL.GL as gl
from OpenGL.arrays import vbo
from OpenGL.GL import shaders
from plots.data import jinja_env
from plots.i18n import _
import math
import numpy as np

from plots import utils
from plots.text import TextRenderer

class GraphArea(Gtk.GLArea):
    __gtype_name__ = "GraphArea"

    INIT_SCALE = 10
    ZOOM_BUTTON_FACTOR = 0.3
    ZOOM_MINIMUM = 5e-34
    ZOOM_MAXIMUM = 1e+38

    def __init__(self):
        super().__init__()
        self.scale = self._target_scale = self.INIT_SCALE
        self._translation = np.array([0, 0], 'f')
        self.vertex_template = jinja_env.get_template('vertex.glsl')
        self.fragment_template = jinja_env.get_template('fragment.glsl')
        self.export_target = None
        self.connect("render", self.gl_render)
        self.connect("realize", self.gl_realize)
        self.app = None

        drag = Gtk.GestureDrag()
        drag.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        drag.connect("drag-update", self.drag_update)
        drag.connect("drag-begin", self.drag_begin)
        self.add_controller(drag)

        scroll_ctl = Gtk.EventControllerScroll()
        scroll_ctl.connect("scroll", self.scroll_zoom)
        scroll_ctl.set_flags(Gtk.EventControllerScrollFlags.VERTICAL)
        self.add_controller(scroll_ctl)

        self.vertex_shader = None

    @property
    def target_scale(self):
        return self._target_scale

    @target_scale.setter
    def target_scale(self, value):
        self._target_scale = value
        self.update_zoom_reset()

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, value):
        self._translation = value
        self.update_zoom_reset()

    def do_css_changed(self, change):
        self.style_cb(self)
        self.queue_draw()

    @staticmethod
    def major_grid(pixel_extent):
        min_extent = 100*pixel_extent
        try:
            exponent = math.floor(math.log10(abs(min_extent)))
        except OverflowError:
            return 0, 0
        mantissa = min_extent/10**exponent
        major = 1.0
        for m in (2.0, 5.0, 10.0):
            if m > mantissa:
                major = m * 10**exponent
                minor = major / (4 if m == 2 else 5)
                return major, minor

    def graph_to_device(self, graph_pos):
        normalised = (graph_pos + self.translation)/self.scale
        gl_pos = normalised / self.viewport * self.viewport[0]
        gl_pos[1] *= -1
        return (gl_pos/2 + 0.5) * self.viewport

    def device_to_graph(self, pixel):
        gl_pos = 2*(pixel/self.viewport - 0.5)
        gl_pos[1] *= -1
        normalised = gl_pos * self.viewport / self.viewport[0]
        return normalised*self.scale - self.translation

    def gl_realize(self, area):
        area.make_current()
        self.style_cb(area)

        if (area.get_error() is not None):
            return

        ctx = self.get_context()
        v = ctx.get_version()
        es = ctx.get_use_es()
        print("OpenGL {}version: {}.{}".format("ES " if es else "", *v))

        version = gl.glGetString(gl.GL_VERSION).decode().split(" ")[0]
        if version < "3.3":
            self.app.errorlabel.set_text(
                _("Warning: OpenGL {} is unsupported. Plots supports OpenGL 3.3 or greater.").format(version))
            self.app.errorbar.props.revealed = True

        self.vertex_shader = shaders.compileShader(
            self.vertex_template.render(), gl.GL_VERTEX_SHADER)
        self.app.update_shader()

        self.vbo = vbo.VBO(np.array([
            [-1, -1, 0],
            [-1, 1, 0],
            [1, 1, 0],
            [-1, -1, 0],
            [1, -1, 0],
            [1, 1, 0]
        ], 'f'), usage="GL_STATIC_DRAW_ARB")

        self.vao = gl.glGenVertexArrays(1)
        gl.glBindVertexArray(self.vao)
        self.vbo.bind()
        self.vbo.copy_data()
        gl.glVertexAttribPointer(0, 3, gl.GL_FLOAT, gl.GL_FALSE, 3*self.vbo.data.itemsize, self.vbo)
        gl.glEnableVertexAttribArray(0)
        self.vbo.unbind()
        gl.glBindVertexArray(0)

        self.text_renderer = TextRenderer(scale_factor=area.get_scale_factor())

    def gl_render(self, area, context):
        area.make_current()
        w = area.get_allocated_width() * area.get_scale_factor()
        h = area.get_allocated_height() * area.get_scale_factor()
        self.viewport = np.array([w, h], 'f')
        self.render()

        if self.export_target:
            width, height = self.viewport.astype(int)
            pixels = gl.glReadPixels(0, 0, width, height, gl.GL_RGB, gl.GL_UNSIGNED_BYTE)
            pixbuf = GdkPixbuf.Pixbuf.new_from_data(
                pixels, GdkPixbuf.Colorspace.RGB, False, 8,
                width, height, width*3, None, None
            ).flip(horizontal=False)
            pixbuf.savev(self.export_target, "png", [])
            self.export_target = None
        return True

    def render(self):
        graph_extent = 2*self.viewport/self.viewport[0]*self.scale
        # extent of each pixel, in graph coordinates
        pixel_extent = graph_extent / self.viewport
        w, h = self.viewport.astype(int)

        gl.glViewport(0, 0, w, h)

        gl.glClearColor(0, 0, 1, 0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glEnable(gl.GL_BLEND)
        gl.glEnable(gl.GL_DEPTH_TEST)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        major_grid, minor_grid = self.major_grid(pixel_extent[0])
        shaders.glUseProgram(self.shader)
        gl.glUniform2f(self.uniform("viewport"), *self.viewport)
        gl.glUniform2f(self.uniform("translation"), *self.translation)
        gl.glUniform2f(self.uniform("pixel_extent"), *pixel_extent)
        gl.glUniform1f(self.uniform("scale"), self.scale)
        gl.glUniform1f(self.uniform("major_grid"), major_grid)
        gl.glUniform1f(self.uniform("minor_grid"), minor_grid)
        gl.glUniform1f(self.uniform("samples"), self.app.prefs["rendering"]["samples"])
        gl.glUniform1f(self.uniform("line_thickness"), self.app.prefs["rendering"]["line_thickness"])
        gl.glUniform3f(self.uniform("fg_color"), *self.fg_color)
        gl.glUniform3f(self.uniform("bg_color"), *self.bg_color)
        for slider in self.app.slider_rows:
            gl.glUniform1f(self.uniform(slider.name), slider.value)
        gl.glBindVertexArray(self.vao)
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, 18)
        gl.glBindVertexArray(0)

        if major_grid <= 0:
            return
        with self.text_renderer.render(w, h) as r:
            low = major_grid * np.floor(
                self.device_to_graph(np.array([0, h]))/major_grid)
            high = major_grid * np.ceil(
                self.device_to_graph(np.array([w, 0]))/major_grid)
            n = (high - low)/major_grid
            pad = 4
            for i in range(round(n[0])+1):
                x = low[0] + i*major_grid
                pos = self.graph_to_device(np.array([x, 0]))
                pos[1] = np.clip(pos[1] + pad, pad, self.viewport[1] - r.top_bearing - pad)
                if x:
                    r.render_text("%g" % x, pos, valign='top', halign='center',
                                  text_color=self.fg_color, bg_color=self.bg_color)
            for j in range(round(n[1])+1):
                y = low[1] + j*major_grid
                label = "%g" % y
                pos = self.graph_to_device(np.array([0, y]))
                pos[0] = np.clip(pos[0] - pad, r.width_of(label) + pad, self.viewport[0] - pad)
                if y:
                    r.render_text(label, pos, valign='center', halign='right',
                                  text_color=self.fg_color, bg_color=self.bg_color)
            r.render_text("0", self.graph_to_device(np.zeros(2)) + np.array([-pad, pad]),
                          valign='top', halign='right', text_color=self.fg_color, bg_color=self.bg_color)


    def style_cb(self, widget):
        ctx = self.get_style_context()
        self.fg_color = utils.rgba_to_tuple(ctx.lookup_color("window_fg_color").color)[:3]
        self.bg_color = utils.rgba_to_tuple(ctx.lookup_color("window_bg_color").color)[:3]

    def uniform(self, name):
        return gl.glGetUniformLocation(self.shader, name)

    def drag_update(self, gesture, dx, dy):
        dr = 2*np.array([dx, -dy], 'f')/self.viewport[0]*self.get_scale_factor()
        self.translation = self.init_translation + dr*self.scale
        self.queue_draw()

    def drag_begin(self, gesture, start_x, start_y):
        self.init_translation = self.translation

    def smooth_scroll(self, translate_to=None):
        speed = 0.3

        if self.target_scale > self.ZOOM_MAXIMUM:
            self.target_scale = self.ZOOM_MAXIMUM
        if self.target_scale < self.ZOOM_MINIMUM:
            self.target_scale = self.ZOOM_MINIMUM

        self.scale = speed*self.target_scale + (1-speed)*self.scale
        if translate_to is not None:
            self.translation = speed*translate_to + (1-speed)*self.translation
        if abs(self.scale/self.target_scale - 1) > 0.01 or \
           (translate_to is not None and
                (np.abs(self.translation - translate_to) > 0.01).any()):
            GLib.timeout_add(1000/60, self.smooth_scroll, translate_to)
        else:
            self.scale = self.target_scale
            if translate_to is not None:
                self.translation = translate_to
        self.queue_draw()

    def scroll_zoom(self, ctl, dx, dy):
        self.target_scale *= math.exp(dy*0.2)
        self.smooth_scroll()

    def zoom(self, button, factor):
        self.target_scale *= math.exp(factor*self.ZOOM_BUTTON_FACTOR)
        self.smooth_scroll()

    def reset_zoom(self, button):
        self.target_scale = self.INIT_SCALE
        if not math.isfinite(self.scale):
            self.scale = self.target_scale
        self.smooth_scroll(translate_to=np.array([0, 0], 'f'))

    def update_zoom_reset(self):
        desired = self.target_scale != self.INIT_SCALE or self.translation.any()
        if self.app.zoom_reset_revealer.get_reveal_child() != desired:
            self.app.zoom_reset_revealer.set_reveal_child(desired)

    def update_fragment_shader(self, formulae):
        if self.vertex_shader:
            self.make_current()
            fragment_shader = shaders.compileShader(
                self.fragment_template.render(formulae=formulae),
                gl.GL_FRAGMENT_SHADER)
            self.shader = shaders.compileProgram(self.vertex_shader, fragment_shader)
            self.queue_draw()



import re
from plots.data import GREEK_REGEXES, FUNCTIONS, BINARY_OPERATORS, GREEK_LETTERS

from . import sum
from . import radical
from . import atom
from . import floor
from . import ceil

def string_to_names(string):
    regex = r"sum|prod|sqrt|nthroot|floor|ceil|."
    regex = "|".join(GREEK_REGEXES) + "|" + "|".join(FUNCTIONS) + "|" + regex
    names = re.findall(regex, string)
    return names

def name_to_element(name):
    if name == 'sqrt':
        return radical.Radical([])
    elif name == 'nthroot':
        return radical.Radical([], index=[])
    elif name == 'sum':
        return sum.Sum()
    elif name == 'prod':
        return sum.Sum(char="∏")
    elif name == 'floor':
        return floor.Floor([])
    elif name == 'ceil':
        return ceil.Ceil([])
    elif name in FUNCTIONS:
        return atom.OperatorAtom(name)
    elif name in BINARY_OPERATORS:
        return atom.BinaryOperatorAtom(name)
    elif len(name) == 1:
        return atom.Atom(name)
    elif name in GREEK_LETTERS:
        return atom.Atom(GREEK_LETTERS[name])
    else:
        return atom.OperatorAtom(name)



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import re
from itertools import count

import gi
gi.require_version('PangoCairo', '1.0')
from gi.repository import GLib, Gtk, Gdk, cairo, Pango, PangoCairo, GObject
from enum import Enum
from plots import parser
from plots.elements import Element, ElementList, BaseAtom, Atom, \
    BinaryOperatorAtom, OperatorAtom, SuperscriptSubscript, Frac, Radical, \
    Abs, Paren, Sum
from plots.utils import Direction, MetricContext, Text
from plots.data import BINARY_OPERATORS, SUP_ATOMS, SUP_TRAFO
from plots.cursor import Cursor

DEBUG = False

class Editor(Gtk.DrawingArea):
    padding = 4
    __gsignals__ = {
        'edit': (GObject.SignalFlags.RUN_LAST, None, ()),
        'cursor_position': (GObject.SignalFlags.RUN_FIRST, None, (float, float)),
    }
    def __init__(self, expression=None):
        super().__init__()
        self.cursor = Cursor(self)
        if expression:
            self.expr = expression
        else:
            self.expr = ElementList()
        self.cursor.reparent(self.expr, -1)
        self.props.can_focus = True
        self.key_ctl = Gtk.EventControllerKey()
        self.key_ctl.connect("key-pressed", self.on_key_press)
        self.add_controller(self.key_ctl)
        self.set_draw_func(self.do_draw_cb)
        click_ctl = Gtk.GestureClick()
        click_ctl.connect("pressed", self.on_button_press)
        self.add_controller(click_ctl)
        self.connect("realize", self.on_realise)
        self.drag_ctl = Gtk.GestureDrag()
        self.drag_ctl.connect("drag-begin", self.on_drag_begin)
        self.drag_ctl.connect("drag-update", self.on_pointer_move)
        self.add_controller(self.drag_ctl)
        focus_ctl = Gtk.EventControllerFocus()
        focus_ctl.connect("enter", self.focus_in)
        focus_ctl.connect("leave", self.focus_out)
        self.add_controller(focus_ctl)
        self.blink_source = None
        self.restart_blink_sequence()
        self.set_size_request(16, 20)
        self.set_focusable(True)

    def set_expr(self, new_expr):
        self.expr = new_expr
        self.cursor.reparent(self.expr, -1)
        self.cursor.cancel_selection()

    def do_draw_cb(self, widget, ctx, w, h):
        Element.color = self.get_style_context().get_color()
        widget_transform = ctx.get_matrix()
        widget_transform.invert()
        ctx.translate(self.padding, self.padding) # a bit of padding
        scale = 1
        ctx.scale(scale, scale)
        self.expr.compute_metrics(ctx, MetricContext(self.cursor))
        ctx.translate(0, self.expr.ascent)
        self.set_size_request(self.expr.width*scale + 2*self.padding,
                              (self.expr.ascent + self.expr.descent)*scale + 2*self.padding)
        self.expr.draw(ctx, self.cursor, widget_transform)
        if self.cursor.position_changed:
            self.emit("cursor_position", *self.cursor.position)
            self.cursor.position_changed = False

    def focus_in(self, ctl):
        self.restart_blink_sequence()

    def focus_out(self, ctl):
        GLib.source_remove(self.blink_source)
        self.blink_source = None
        self.cursor.cancel_selection()
        self.cursor.visible = False
        self.queue_draw()

    def blink_cursor_cb(self):
        self.cursor.visible = not self.cursor.visible
        self.queue_draw()
        return True

    def restart_blink_sequence(self):
        if not self.cursor.visible:
            self.cursor.visible = True
        if self.blink_source:
            GLib.source_remove(self.blink_source)
        self.blink_source = GLib.timeout_add(Cursor.BLINK_DELAY, self.blink_cursor_cb)

    def on_key_press(self, event_cont, keyval, keycode, state):
        self.restart_blink_sequence()
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        select = bool(modifiers & Gdk.ModifierType.SHIFT_MASK)
        if DEBUG:
            print(Gdk.keyval_name(keyval))
        if modifiers & (Gdk.ModifierType.ALT_MASK | Gdk.ModifierType.SUPER_MASK):
            return False
        try:
            direction = Direction(keyval)
            res = self.cursor.handle_movement(direction, select=select)
            self.queue_draw()
            return res
        except ValueError:
            pass
        if keyval in (Gdk.KEY_Home, Gdk.KEY_End):
            if not select:
                self.cursor.cancel_selection()
            if select and not self.cursor.selecting:
                self.cursor.secondary_pos = self.cursor.pos
                self.cursor.secondary_owner = self.cursor.owner
                self.cursor.selecting = True
            new_pos = 0 if keyval == Gdk.KEY_Home else -1
            self.cursor.reparent(self.expr, new_pos)
            if select:
               self.cursor.selection_bounds, self.cursor.selection_ancestor = self.cursor.calculate_selection()
            self.queue_draw()
            return True

        char = chr(Gdk.keyval_to_unicode(keyval))
        if modifiers & Gdk.ModifierType.CONTROL_MASK:
            if char == "a":
                self.cursor.select_all(self.expr)
                self.queue_draw()
                return True
            elif char == "c":
                self.cursor.copy_selection()
                self.queue_draw()
                return True
            elif char == "x":
                self.cursor.cut_selection()
                self.queue_draw()
                self.emit("edit")
                return True
            elif char == "v":
                self.cursor.paste()
                # cursor will emit("edit")
                return True
            else:
                return False
        if char in SUP_ATOMS:
            self.cursor.insert_superscript_subscript(superscript=True)
            self.cursor.insert(Atom(char.translate(SUP_TRAFO)))
            self.cursor.handle_movement(Direction(Gdk.KEY_Right), select=False) # reset cursor level
            self.queue_draw()
            self.emit("edit")
            return
        if char.isalnum():
            self.cursor.insert(Atom(char))
            self.queue_draw()
            self.emit("edit")
            return
        if char in BINARY_OPERATORS:
            translation = str.maketrans("-*", "−×")
            self.cursor.insert(BinaryOperatorAtom(char.translate(translation)))
            self.queue_draw()
            self.emit("edit")
            return
        if char in "!'.":
            translation = str.maketrans("'", "′")
            self.cursor.insert(Atom(char.translate(translation)))
            self.queue_draw()
            self.emit("edit")
            return
        if char in "()[]{}":
            self.cursor.insert(Paren(char))
            self.queue_draw()
            self.emit("edit")
            return
        if keyval == Gdk.KEY_BackSpace:
            self.cursor.backspace(Direction.LEFT)
            self.queue_draw()
            self.emit("edit")
            return
        if keyval == Gdk.KEY_Delete:
            self.cursor.backspace(Direction.RIGHT)
            self.queue_draw()
            self.emit("edit")
            return
        if keyval in (Gdk.KEY_slash, Gdk.KEY_KP_Divide):
            self.cursor.greedy_insert(Frac)
            self.queue_draw()
            self.emit("edit")
            return
        if char == "^" or keyval == Gdk.KEY_dead_circumflex:
            self.cursor.insert_superscript_subscript(superscript=True)
            self.queue_draw()
            self.emit("edit")
            return
        if char == "_":
            self.cursor.insert_superscript_subscript(superscript=False)
            self.queue_draw()
            self.emit("edit")
            return
        if char == "|":
            self.cursor.insert(Abs(None))
            self.queue_draw()
            self.emit("edit")
            return

        self.queue_draw()

    def element_at(self, x, y):
        e = self.expr
        while True:
            for c in e.children():
                if c.contains_device_point(x, y):
                    e = c
                    break
            else:
                return e, e.half_containing(x, y)

    def on_button_press(self, ctl, n_press, x, y):
        if n_press > 1:
            self.cursor.select_all(self.expr)
        else:
            element, direction = self.element_at(x, y)
            self.cursor.mouse_select(element, direction, drag=False)
            self.restart_blink_sequence()
            self.grab_focus()
        self.queue_draw()
        return True

    def on_drag_begin(self, ctl, start_x, start_y):
        self.drag_start_x = start_x
        self.drag_start_y = start_y

    def on_pointer_move(self, ctl, x, y):
        element, direction = self.element_at(
            self.drag_start_x + x, self.drag_start_y + y)
        self.cursor.mouse_select(element, direction, drag=True)
        self.restart_blink_sequence()
        self.queue_draw()

    def on_realise(self, widget):
        self.set_cursor(Gdk.Cursor.new_from_name("text", None))



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from plots.cursor import Cursor
import plots.formula as f
from plots.parser import from_latex
from plots.utils import Direction

from tests.fixtures import cursor

import pytest

def test_position_changed():
    cursor = Cursor()
    assert cursor.position_changed == False
    cursor.position = (1., 1.)
    assert cursor.position_changed == True

def test_position_not_changed():
    cursor = Cursor()
    assert cursor.position_changed == False
    cursor.position = (0., 0.)
    assert cursor.position_changed == False

def test_select_all_backspace():
    cursor = Cursor()
    elems = from_latex("abcd")
    cursor.reparent(elems, 2)
    cursor.select_all(elems)
    cursor.backspace(None)
    assert len(elems) == 0

def test_backspace():
    cursor = Cursor()
    elems = from_latex("abcd")
    cursor.reparent(elems, 2)
    cursor.backspace(Direction.LEFT)
    assert elems.to_latex() == "acd"

def test_move_to_next_list(cursor):
    elems = from_latex(r"\frac{3}{2}")
    cursor.reparent(elems[0].numerator, -1)
    cursor.handle_movement(Direction.RIGHT)
    assert cursor.owner is elems[0].denominator
    assert cursor.pos == 0

def test_move_to_previous_list(cursor):
    elems = from_latex(r"\frac{3}{2}")
    cursor.reparent(elems[0].denominator, 0)
    cursor.handle_movement(Direction.LEFT)
    assert cursor.owner is elems[0].numerator
    assert cursor.pos == 1

def test_ancestors(cursor):
    elems = from_latex(r"\sqrt{\operatorname{cos}\frac{3}{4\abs{\sum_{i=1}^{3}}}}")
    expected = [
        elems[0].radicand[1].denominator[1].argument[0].bottom,
        elems[0].radicand[1].denominator[1].argument,
        elems[0].radicand[1].denominator,
        elems[0].radicand,
        elems,
    ]
    actual = cursor.ancestors(elems[0].radicand[1].denominator[1].argument[0].bottom)
    for e, a in zip(expected, actual):
        assert e is a

def test_calculate_selection_deep(cursor):
    elems = from_latex(r"x\abs{3}-3\sqrt{\frac{2}{x}}+4")
    cursor.reparent(elems[1].argument, -1)
    cursor.secondary_owner = elems[4].radicand[0].denominator
    cursor.secondary_pos = 0
    _range, ancestor = cursor.calculate_selection()
    assert _range == range(1, 5)
    assert ancestor is elems

def test_calculate_selection_deep_alternative(cursor):
    elems = from_latex(r"x\abs{3}-3\sqrt{\frac{2}{x}}+4")
    cursor.reparent(elems[4].radicand[0].denominator, -1)
    cursor.secondary_owner = elems[1].argument
    cursor.secondary_pos = 0
    _range, ancestor = cursor.calculate_selection()
    assert _range == range(1, 5)
    assert ancestor is elems

def test_mouse_select_drag_deep(cursor):
    elems = from_latex(r"x\abs{3}-3\sqrt{\frac{2}{x}}+4")
    cursor.reparent(elems[4].radicand[0].denominator, -1)
    cursor.mouse_select(elems[1].argument[0], Direction.LEFT, drag=True)
    assert cursor.selection_bounds == range(1, 5)
    assert cursor.selection_ancestor == elems

def test_mouse_select_left(cursor):
    elems = from_latex(r"abcd")
    cursor.mouse_select(elems[1], Direction.LEFT, drag=False)
    assert cursor.owner is elems
    assert cursor.pos is 1

def test_mouse_select_right(cursor):
    elems = from_latex(r"abcd")
    cursor.mouse_select(elems[1], Direction.RIGHT, drag=False)
    assert cursor.owner is elems
    assert cursor.pos is 2

def test_mouse_select_list(cursor):
    elems = from_latex(r"3x\abs{}")
    cursor.mouse_select(elems[2].argument, Direction.RIGHT, drag=False)
    assert cursor.owner is elems[2].argument
    assert cursor.pos is 0

def test_mouse_select_then_drag(cursor):
    elems = from_latex(r"abcd")
    cursor.mouse_select(elems[1], Direction.LEFT, drag=False)
    cursor.mouse_select(elems[3], Direction.LEFT, drag=True)
    assert cursor.owner is elems
    assert cursor.selection_bounds == range(1, 3)

def test_handle_movement_up_out(cursor):
    elems = from_latex(r"\abs{x}")
    cursor.reparent(elems[0].argument, -1)
    cursor.handle_movement(Direction.UP)
    assert cursor.owner is elems
    assert cursor.pos == 0

def test_handle_movement_down_out(cursor):
    elems = from_latex(r"\abs{x}")
    cursor.reparent(elems[0].argument, -1)
    cursor.handle_movement(Direction.DOWN)
    assert cursor.owner is elems
    assert cursor.pos == 1



from . import element
from plots.utils import saved, italify_string, deitalify_string, Text
from plots.data import GREEK_LETTERS_INVERSE

class BaseAtom(element.Element):
    h_spacing = 0

    def __init__(self, name, parent=None):
        super().__init__(parent)
        self.name = name

    def compute_metrics(self, ctx, metric_ctx):
        self.layout = Text(self.name, ctx)
        self.width, self.ascent, self.descent = self.layout.width, self.layout.ascent, self.layout.descent
        super().compute_metrics(ctx, metric_ctx)

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        self.layout.draw_at_baseline(ctx)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.name)

    def __eq__(self, other):
        if isinstance(other, BaseAtom):
            return self.name == other.name
        return NotImplemented

    def to_glsl(self):
        s = deitalify_string(self.name)
        if s in GREEK_LETTERS_INVERSE:
            return "", GREEK_LETTERS_INVERSE[s]
        else:
            return "", deitalify_string(self.name)

    def to_latex(self):
        return deitalify_string(self.name)

class Atom(BaseAtom):
    def __init__(self, name, parent=None):
        super().__init__(italify_string(name), parent=parent)

    @staticmethod
    def part_of_number(element):
        return isinstance(element, Atom) \
            and (element.name.isdigit() or element.name == ".")

class BinaryOperatorAtom(BaseAtom):
    def __init__(self, name, parent=None):
        super().__init__(name, parent=parent)
        if name == "=":
            self.h_spacing = 6
        else:
            self.h_spacing = 4

    def to_glsl(self):
        translation = str.maketrans("−×", "-*")
        return "", self.name.translate(translation)

    def to_latex(self):
        if self.name == "−":
            return "-"
        elif self.name == "×":
            return "\\times "
        else:
            return self.name

class OperatorAtom(BaseAtom):
    h_spacing = 2

    def to_latex(self):
        return "\\operatorname{" + self.name + "}"



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import plots.elements as e
import plots.parser

@pytest.mark.parametrize("elems, expected", [
    ([e.Atom('x')],
     'x'),
    ([e.Atom('x'), e.Atom('y')],
     'x*y'),
    ([e.Frac([e.Atom('x')], [e.Atom('y')])],
     '(x)/(y)'),
    ([e.Abs([e.Atom('x')])],
     r'abs(x)'),
    ([e.Floor([e.Atom('x')]), e.Atom('y')],
     r'floor(x)*y'),
    ([e.Ceil([e.Atom('x')]), e.SuperscriptSubscript(exponent=e.ElementList([e.Atom('3')]))],
     r'mypow(ceil(x), (3.0))'),
    ([e.Atom('x'), e.SuperscriptSubscript(
        exponent=e.ElementList([e.Atom('2')]))],
     'mypow(x, (2.0))'),
    ([e.Atom('x'), e.Atom('!')],
     'factorial(x)'),
    ([e.OperatorAtom('sin'), e.Atom('x')],
     'sin(x)'),
    ([e.OperatorAtom('log'), e.Atom('x')],
     'log(x)'),
    ([e.OperatorAtom('ln'), e.Atom('x')],
     'ln(x)'),
    ([e.OperatorAtom('lg'), e.Atom('x')],
     'lg(x)'),
    ([e.OperatorAtom('log'), e.SuperscriptSubscript(subscript=e.ElementList([e.Atom('8')])), e.Atom('x')],
     'log_base(8.0,x)'),
    ([e.OperatorAtom('sin'), e.Atom('3'), e.Atom('x')],
     'sin(3.0*x)'),
    ([e.OperatorAtom('sin'), e.Paren("{"), e.Atom('3'), e.Atom('x'), e.Paren("}"), e.Atom('y')],
     'sin(3.0*x)*y'),
    ([e.OperatorAtom('sin'), e.Paren("{"), e.Atom('3'), e.Atom('x'), e.Paren("}"),
      e.SuperscriptSubscript(
          exponent=e.ElementList([e.Atom('3'), e.Atom('z')])
      )],
     'mypow(sin(3.0*x), (3.0*z))'),
    ([e.OperatorAtom('sin'), e.Atom('3'), e.BinaryOperatorAtom('-'), e.Atom('x')],
     'sin(3.0)-x'),
    ([e.OperatorAtom('sin'), e.BinaryOperatorAtom('-'), e.Atom('x')],
     'sin(-x)'),
    ([e.Atom('y'), e.Radical([
        e.Atom('x'), e.BinaryOperatorAtom('-'), e.Atom('1')])],
     'y*sqrt(x-1.0)'),
    ([e.Atom('y'), e.Radical([
        e.Atom('x'), e.BinaryOperatorAtom('-'), e.Atom('1')],
                             index=[e.Atom('3')])],
     'y*pow(x-1.0, 1.0/(3.0))'),
])
def test_elementlist_to_glsl(elems, expected):
    assert e.ElementList(elems).to_glsl() == ('', expected)

@pytest.mark.parametrize('string', [
    "x",
    "3x",
    "xyz^{4}",
    r"\operatorname{sin}\frac{3}{x}",
    r"\abs{y^{2}}^{2}",
    r"3\floor{\frac{1}{x}}",
    r"3\ceil{\frac{1}{x}}",
    r"ΑαΒβΓγΔδΕεΖζΗηΘθΙιΚκΛλΜμΝνΞξΟοΠπΡρΣσςΤτΥυΦφΧχΨψΩω",
    "33!",
    r"\operatorname{log}_{10}^{2}(x^{3})",
    r"\sqrt{9x_{a}}",
    r"\sqrt[3-n]{x}",
    r"\sum_{i=1}^{n}[36i^{2}]",
    r"\prod_{a=-5}^{-1}\{abc\}",
    r"-3.2342838^{5}",
    r"+-\times=",
    r"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
])
def test_parse_and_serialize(string):
    assert string == plots.parser.from_latex(string).to_latex()

def test_parse_parent_pointers():
    e = plots.parser.from_latex(r"\sqrt{\frac{x}{2}}")
    assert e[0].parent is e
    assert e[0].radicand.parent is e[0]
    r = e[0].radicand
    assert r[0].parent is r
    n, d = r[0].numerator, r[0].denominator
    assert n.parent is d.parent is r[0]
    assert n[0].parent is n
    assert d[0].parent is d



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.


import pytest
import re

import plots.formula as f
from plots.parser import from_latex

from tests.fixtures import cursor

def clean(s):
    return re.sub(r" +", " ", s.replace("\n", " ")).strip()

@pytest.mark.parametrize('latex, body, expr', [
    (r"\sum_{i=1}^{50}3i-1", """
float sum0 = 0.0;
for (float i=1.0; i <= 50.0; i++) {
    sum0 += 3.0*i;
}
""", "sum0-1.0"),
    # x + x Σ[x^2(Σ x^{ij}) + Σexp(ikx)]
    (r"x + x\sum_{i=1}^{4}[x^{2}(\sum_{j=1}^{i}x^{ij})+\sum_{k=1}^{i^{2}}\operatorname{exp}(ikx)]", """
float sum2 = 0.0;
for (float i=1.0; i <= 4.0; i++) {
    float sum0 = 0.0;
    for (float j=1.0; j <= i; j++) {
        sum0 += mypow(x, (i*j));
    }
    float sum1 = 0.0;
    for (float k=1.0; k <= mypow(i, (2.0)); k++) {
        sum1 += exp(i*k*x);
    }
    sum2 += (mypow(x, (2.0))*(sum0)+sum1);
}
""", "x+x*sum2"),
    (r"\prod_{i=4}^{50}\operatorname{sin}(3ix-1)", """
float sum0 = 1.0;
for (float i=4.0; i <= 50.0; i++) {
    sum0 *= sin(3.0*i*x-1.0);
}
""", "sum0"),
])
def test_sum_to_glsl(latex, body, expr):
    f.Sum.glsl_var_counter = 0
    glsl = from_latex(latex).to_glsl()
    assert clean(glsl[0]) == clean(body)
    assert glsl[1] == expr



import gi
from gi.repository import GLib, Gtk, Gdk

from plots.utils import saved, Direction

from . import elements
from . import abstractelement

DEBUG = False

class Element(abstractelement.AbstractElement):
    """Abstract class describing an element of an equation.

    Implementations must provide parent, index_in_parent, lists, ascent, descent,
    and width properties, compute_metrics(ctx, metric_ctx) and draw(ctx, cursor, widget_transform)."""

    h_spacing = 2
    color = Gdk.RGBA()

    def __init__(self, parent):
        super().__init__(parent)
        self.index_in_parent = None
        self.lists = []
        self.default_list = None
        self.cursor_acceptor = None

    def children(self):
        return self.lists

    def compute_metrics(self, ctx, metric_ctx):
        """To be run at the end of overriding methods, if they
        wish to have parens scale around them.
        """
        stack = metric_ctx.paren_stack
        if stack:
            stack[-1].ascent = max(self.ascent, stack[-1].ascent)
            stack[-1].descent = max(self.descent, stack[-1].descent)
            stack[-1].compute_stretch()

    def draw(self, ctx, cursor, widget_transform):
        """Expects (0, 0) to be at the baseline, where it should begin drawing.
        """
        super().draw(ctx, cursor, widget_transform)
        if DEBUG:
            ctx.set_line_width(0.5)
            ctx.set_source_rgba(1, 0, 1 if cursor.owner is self else 0, 0.6)
            ctx.rectangle(0, -self.ascent, self.width, self.ascent + self.descent)
            ctx.stroke()
        if cursor.selecting and self.parent is cursor.selection_ancestor and \
           self.index_in_parent in cursor.selection_bounds:
            ctx.set_source_rgba(*cursor.selection_rgba)
            ctx.rectangle(-self.h_spacing, -self.ascent,
                          self.width + 2*self.h_spacing, self.ascent + self.descent)
            ctx.fill()
        Gdk.cairo_set_source_rgba(ctx, Element.color)
        ctx.move_to(0, 0)

    def get_next_child(self, direction, previous=None):
        try:
            previous_idx = self.lists.index(previous)
            new_idx = previous_idx + direction.displacement()
            if new_idx in range(len(self.lists)):
                return self.lists[new_idx]
            else:
                return None
        except ValueError:
            child_idx = -1 if direction.displacement() == -1 else 0
            if self.lists:
                return self.lists[child_idx]
            else:
                return None

    def accept_selection(self, elements, direction):
        pass

    def dissolve(self, cursor, caller):
        concatenation = []
        cursor_offset = 0
        for elementlist in self.lists:
            if elementlist is caller:
                cursor_offset = len(concatenation)
            concatenation.extend(elementlist.elements)
        self.parent.replace(self, elements.ElementList(concatenation), cursor, cursor_offset)



import setuptools

package = 'Plots'
version = '0.8.5'

setuptools.setup(
    name=package,
    version=version,
    description="A graph plotting app for GNOME",
    packages=setuptools.find_packages(exclude=['tests']),
    package_data={
        "plots.ui": ["*.ui"],
        "plots": ["shaders/*.glsl"],
        "plots.res": ["*.svg", "*.ttf"],
        "plots.locale": ["*/LC_MESSAGES/*.mo"],
    },
    install_requires=[
        "PyGObject",
        "PyOpenGL",
        "Jinja2",
        "numpy",
        "lark",
        "PyGLM",
        "freetype-py",
    ],
    python_requires='~=3.6',
    entry_points={
        "gui_scripts": [
            "plots=plots:main"
        ],
    },
)



from . import elements
from . import element
from plots.utils import saved, Text

class Paren(element.Element):
    h_spacing = 0
    shrink = 0.7

    def __init__(self, char, parent=None):
        super().__init__(parent)
        if len(char) != 1:
            raise ValueError("{!r} is not a valid paren".format(char))
        if char in "({[":
            self.left = True
        elif char in "]})":
            self.left = False
        else:
            raise ValueError("{!r} is not a valid paren".format(char))
        self.char = char
        self.match = None

    def __repr__(self):
        return f'Paren({self.char!r})'

    def compute_metrics(self, ctx, metric_ctx):
        self.text = Text(self.char, ctx)
        if self.char == "[":
            self.top, self.mid, self.bot = [Text(c, ctx) for c in "⎡⎢⎣"]
        elif self.char == "]":
            self.top, self.mid, self.bot = [Text(c, ctx) for c in "⎤⎥⎦"]

        self.width, self.ascent, self.descent = self.text.width, self.text.ascent, self.text.descent

        if self.left:
            metric_ctx.paren_stack.append(self)
        else:
            if metric_ctx.paren_stack:
                self.match = metric_ctx.paren_stack.pop()
            else:
                self.match = metric_ctx.prev
            self.ascent = self.match.ascent
            self.descent = self.match.descent
            super().compute_metrics(ctx, metric_ctx)
        self.compute_stretch()

    def compute_stretch(self):
        self.scale_factor = max(1, (self.ascent + self.descent)/self.text.ink_rect.height)
        if self.scale_factor > 1.5 and self.char in "[]":
            self.stretch = True
            self.scale_factor = max(1, (self.ascent + self.descent)/self.mid.height)
            self.width = self.mid.width*self.shrink
            self.h_spacing = 0
            if isinstance(self.match, Paren) and self.match.char in "[]":
                self.match.stretch = True
                self.match.scale_factor = self.scale_factor
                self.match.width = self.width
                self.match.h_spacing = self.h_spacing
        else:
            self.stretch = False

    def draw(self, ctx, cursor, widget_transform):
        super().draw(ctx, cursor, widget_transform)
        if self.stretch:
            with saved(ctx):
                ctx.translate(0, -self.ascent - self.top.ink_rect.y*self.shrink)
                ctx.move_to(0,0)
                ctx.scale(self.shrink,self.shrink)
                self.top.draw(ctx)
            with saved(ctx):
                ctx.translate(0, self.descent)
                ctx.move_to(0,0)
                ctx.scale(self.shrink,self.shrink)
                ctx.translate(0, -self.bot.ink_rect.y - self.bot.ink_rect.height)
                ctx.move_to(0,0)
                self.bot.draw(ctx)
            with saved(ctx):
                scale_factor = max(1, (self.ascent + self.descent)/self.mid.ink_rect.height)
                ctx.translate(0, -self.ascent)
                ctx.scale(1, self.scale_factor)
                ctx.translate(0, -self.mid.ink_rect.y)
                ctx.scale(self.shrink,1)
                ctx.move_to(0, 0)
                self.mid.draw(ctx)
        else:
            with saved(ctx):
                ctx.scale(1, self.scale_factor)
                ctx.translate(0, -self.ascent/self.scale_factor-self.text.ink_rect.y)
                ctx.move_to(0, 0)
                self.text.draw(ctx)

    def to_glsl(self):
        return "", "(" if self.left else ")"

    def to_latex(self):
        if self.char in "{}":
            return "\\" + self.char
        else:
            return self.char

    @classmethod
    def is_paren(cls, element, left=None):
        if not isinstance(element, cls):
            return False
        if left is None:
            return True
        return left == element.left



# Copyright 2021 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import plots.elements as e
from plots.parser import from_latex
from plots.cursor import Cursor
import plots.data
import plots.utils

from tests.fixtures import cursor

def do_convert_specials(name):
    elems = from_latex(name)
    cursor = Cursor()
    cursor.reparent(elems, -1)
    elems.convert_specials(cursor)
    return elems

@pytest.mark.parametrize("fn", plots.data.FUNCTIONS)
def test_convert_specials_functions(fn):
    elems = do_convert_specials(fn)
    assert len(elems) == 1
    elem = elems[0]
    assert isinstance(elem, e.OperatorAtom) and elem.name == fn

@pytest.mark.parametrize("name, symbol", [tup for tup in plots.data.GREEK_LETTERS.items()])
def test_convert_specials_greek(name, symbol):
    elems = do_convert_specials(name)
    assert len(elems) == 1
    elem = elems[0]
    assert isinstance(elem, e.Atom) \
        and plots.utils.deitalify_string(elem.name) == symbol

@pytest.mark.parametrize("name", "epsi upsi".split())
def test_convert_specials_epsilon_upsilon(name):
    elems = do_convert_specials(name)
    assert len(elems) == len(name)
    assert elems.to_latex() == name

@pytest.mark.parametrize("name, cls", [
    ("sum", e.Sum),
    ("prod", e.Sum),
    ("sqrt", e.Radical),
    ("nthroot", e.Radical),
    ("floor", e.Floor),
    ("ceil", e.Ceil),
])
def test_convert_specials_misc(name, cls):
    elems = do_convert_specials(name)
    assert len(elems) == 1
    assert isinstance(elems[0], cls)

@pytest.mark.parametrize("base", "sin cos tan sec cosec csc cot".split())
@pytest.mark.parametrize("template", ["a{}", "arc{}", "{}h", "a{}h"])
def test_convert_specials_add_to_trig(base, template):
    op = f"\\operatorname{{{base}}}"
    elems = do_convert_specials(template.format(op))
    assert len(elems) == 1
    assert elems[0].name == template.format(base)

def test_greedy_insert(cursor):
    elems = from_latex(r"3\sqrt{x}(x(9-x))-4")
    cursor.reparent(elems, 2)
    cursor.greedy_insert(e.Frac)
    assert elems.to_latex() == r"3\frac{\sqrt{x}}{(x(9-x))}-4"

def test_greedy_insert_with_parens(cursor):
    elems = from_latex(r"3(x+1)(x(9-x))-4")
    cursor.reparent(elems, 6)
    cursor.greedy_insert(e.Frac)
    assert elems.to_latex() == r"3\frac{(x+1)}{(x(9-x))}-4"

def test_greedy_insert_with_numbers(cursor):
    elems = from_latex(r"3.238923829-4")
    cursor.reparent(elems, 6)
    cursor.greedy_insert(e.Frac)
    assert elems.to_latex() == r"\frac{3.2389}{23829}-4"

def test_dissolve(cursor):
    elems = from_latex(r"\frac{abc}{def}")
    cursor.reparent(elems[0].numerator, 0)
    cursor.backspace(plots.utils.Direction.LEFT)
    assert elems.to_latex() == "abcdef"
    assert cursor.owner is elems

def test_backspace_into(cursor):
    elems = from_latex(r"\frac{abc}{def}")
    cursor.reparent(elems, -1)
    cursor.backspace(plots.utils.Direction.LEFT)
    assert elems.to_latex() == r"\frac{abc}{de}"
    assert cursor.owner is elems[0].denominator
    assert cursor.pos == 2

def test_delete_into(cursor):
    elems = from_latex(r"\frac{abc}{def}")
    cursor.reparent(elems, 0)
    cursor.backspace(plots.utils.Direction.RIGHT)
    assert elems.to_latex() == r"\frac{bc}{def}"
    assert cursor.owner is elems[0].numerator
    assert cursor.pos == 0

def test_frac_accept_selection(cursor):
    elems = from_latex(r"xyz")
    cursor.select_all(elems)
    cursor.greedy_insert(e.Frac)
    assert elems.to_latex() == r"\frac{xyz}{}"
    assert cursor.owner is elems[0].denominator



#!/usr/bin/env python3

# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
from gi.repository import Gtk, Gdk, GLib, Gio, GdkPixbuf, Adw

from plots import formula, formularow, rowcommands, preferences, utils, graph
from plots.i18n import _
import plots.i18n
import sys
import importlib.resources as resources

class Plots(Adw.Application):
    def __init__(self):
        super().__init__(application_id="com.github.alexhuntley.Plots")
        self.rows = []
        self.slider_rows = []
        self.history = []
        self.history_position = 0  # index of the last undone command / next in line for redo
        self.overlay_source = None
        Adw.StyleManager.get_default().set_color_scheme(
            Adw.ColorScheme.PREFER_LIGHT)
        utils.install_excepthook()
        plots.i18n.bind()

    def key_pressed(self, ctl, keyval, keycode, state):
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        char = chr(Gdk.keyval_to_unicode(keyval))
        if keyval == Gdk.KEY_Return:
            self.add_equation(None)
            return True
        elif modifiers & Gdk.ModifierType.CONTROL_MASK:
            if char == "z":
                self.undo(None)
                return True
            elif char == "y" or char == "Z" and modifiers & Gdk.ModifierType.SHIFT_MASK:
                self.redo(None)
                return True

    def do_activate(self):
        builder = Gtk.Builder()
        builder.add_from_string(utils.read_ui_file("plots.ui"))

        self.window = builder.get_object("main_window")
        self.add_window(self.window)
        self.window.set_title(_("Plots"))
        self.scroll = builder.get_object("equation_scroll")
        self.formula_box = builder.get_object("equation_box")
        self.add_equation_button = builder.get_object("add_equation")
        self.undo_button = builder.get_object("undo")
        self.redo_button = builder.get_object("redo")
        self.key_ctl = Gtk.EventControllerKey()
        self.key_ctl.connect("key-pressed", self.key_pressed)
        self.window.add_controller(self.key_ctl)
        self.window.connect("close-request", self.delete_cb)

        self.gl_area = builder.get_object("gl")
        self.gl_area.app = self

        self.errorbar = builder.get_object("errorbar")
        self.errorbar.set_message_type(Gtk.MessageType.ERROR)
        self.errorbar.connect("response", lambda id, data: self.errorbar.set_property("revealed", False))
        self.errorbar.props.revealed = False
        self.errorlabel = builder.get_object("errorlabel")

        self.add_equation_button.connect("clicked", self.add_equation)
        self.undo_button.connect("clicked", self.undo)
        self.redo_button.connect("clicked", self.redo)

        self.osd_revealer = builder.get_object("osd_revealer")
        self.osd_box = builder.get_object("osd_box")
        self.zoom_reset_revealer = builder.get_object("zoom_reset_revealer")
        self.graph_overlay = builder.get_object("graph_overlay")
        self.zoom_in_button = builder.get_object("zoom_in")
        self.zoom_in_button.connect("clicked", self.gl_area.zoom, -1)
        self.zoom_out_button = builder.get_object("zoom_out")
        self.zoom_out_button.connect("clicked", self.gl_area.zoom, 1)
        self.zoom_reset_button = builder.get_object("zoom_reset")
        self.zoom_reset_button.connect("clicked", self.gl_area.reset_zoom)
        self.gl_area.update_zoom_reset()

        menu_button = builder.get_object("menu_button")

        self.menu = Gio.Menu()
        self.menu.append(_("_Export…"), "app.export")
        self.menu.append(_("_Preferences"), "app.preferences")
        self.menu.append(_("Help"), "app.help")
        self.menu.append(_("About Plots"), "app.about")
        menu_button.set_menu_model(self.menu)

        self.about_action = Gio.SimpleAction.new("about", None)
        self.about_action.connect("activate", self.about_cb)
        self.about_action.set_enabled(True)
        self.add_action(self.about_action)

        about_builder = Gtk.Builder()
        about_builder.add_from_string(utils.read_ui_file("about.ui"))
        self.about_dialog = about_builder.get_object("about_dialog")
        self.about_dialog.props.modal = True
        self.about_dialog.set_transient_for(self.window)
        self.about_dialog.set_modal(True)
        self.about_dialog.connect("close-request", self.about_close)

        with resources.path("plots.res", "com.github.alexhuntley.Plotter.svg") as p:
            texture = Gdk.Texture.new_from_filename(str(p))
            self.about_dialog.set_logo(texture)

        help_action = Gio.SimpleAction.new("help", None)
        help_action.connect("activate", self.help_cb)
        help_action.set_enabled(True)
        self.add_action(help_action)

        export_action = Gio.SimpleAction.new("export", None)
        export_action.connect("activate", self.export_cb)
        export_action.set_enabled(True)
        self.add_action(export_action)

        prefs_action = Gio.SimpleAction.new("preferences", None)
        prefs_action.connect("activate", self.prefs_cb)
        prefs_action.set_enabled(True)
        self.add_action(prefs_action)
        self.prefs = preferences.Preferences(self.window)
        self.prefs.connect("updated", self.prefs_updated)

        self.set_overlay_timeout()

        self.add_equation(None, record=False)

        self.window.set_default_size(1280, 720)

        css = '''
.formula_box {
        background-color: @theme_base_color;
        border-bottom-color: @borders;
        border-bottom-width: 1px;
        border-bottom-style: solid;
}
.zoom-box {
        background-color: rgba(0, 0, 0, 0);
}
.zoom-button {
        padding: 4px;
}
'''
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(css, len(css))
        context = self.window.get_style_context()
        display = self.window.get_display()
        context.add_provider_for_display(display, css_provider,
                                         Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        motion_ctl = Gtk.EventControllerMotion()
        motion_ctl.connect("motion", self.motion_cb)
        self._old_motion = None
        self.gl_area.add_controller(motion_ctl)
        overlay_motion_ctl = Gtk.EventControllerMotion()
        overlay_motion_ctl.connect("enter", self.enter_overlay_cb)
        self.osd_box.add_controller(overlay_motion_ctl)
        self.refresh_history_buttons()
        self.window.show()

    def clear_overlay_timeout(self):
        if self.overlay_source is not None:
            GLib.source_remove(self.overlay_source)
            self.overlay_source = None

    def set_overlay_timeout(self):
        self.clear_overlay_timeout()
        self.overlay_source = GLib.timeout_add(2000, self.overlay_timeout_cb)

    def overlay_timeout_cb(self):
        self.osd_revealer.set_reveal_child(False)
        self.overlay_source = None

    def motion_cb(self, ctl, x, y):
        if (x, y) == self._old_motion:
            return False
        if not self.osd_revealer.get_reveal_child():
            self.osd_revealer.set_reveal_child(True)
        self.set_overlay_timeout()
        self._old_motion = (x, y)
        return False

    def enter_overlay_cb(self, ctl, x, y):
        self.clear_overlay_timeout()
        return False

    def update_shader(self):
        good, bad, unknown = [], [], []
        self.slider_rows.clear()
        for r in self.rows:
            data = r.get_data()
            if r.row_status == formularow.RowStatus.GOOD:
                good.append(data)
            elif r.row_status == formularow.RowStatus.BAD:
                bad.append(data)
            elif r.row_status == formularow.RowStatus.UNKNOWN:
                unknown.append(data)
            if isinstance(data, formularow.Slider):
                self.slider_rows.append(r)

        def attempt(formulae):
            formulae.sort(key=lambda x: x.priority, reverse=True)
            self.gl_area.update_fragment_shader(formulae)
            for f in formulae:
                f.owner.row_status = formularow.RowStatus.GOOD

        try:
            attempt(good + bad + unknown)
        except RuntimeError:
            try:
                attempt(good + unknown)
            except RuntimeError:
                try:
                    attempt(good)
                except RuntimeError:
                    attempt([])

    def dependency_changed(self, row):
        for r in self.rows:
            r.row_status = formularow.RowStatus.UNKNOWN

    def add_equation(self, _, record=True):
        row = formularow.FormulaBox(self)
        row.connect("dependency_changed", self.dependency_changed)
        self.rows.append(row)
        self.formula_box.append(row)
        row.editor.grab_focus()
        if record:
            self.add_to_history(rowcommands.Add(row, self.rows))

    def insert_row(self, index, row):
        self.rows.insert(index, row)
        prev = self.rows[index-1] if index > 0 else None
        self.formula_box.insert_child_after(row, prev)
        row.editor.grab_focus()

    def about_cb(self, action, _):
        self.about_dialog.present()

    def about_close(self, window):
        window.hide()
        return True

    def prefs_cb(self, action, param):
        self.prefs.show()

    def help_cb(self, action, _):
        Gtk.show_uri(None, "help:plots", Gdk.CURRENT_TIME)

    def export_cb(self, action, parameter):
        self.export_dialog = Gtk.FileChooserNative.new(
            title=_("Export image"),
            parent=self.window,
            action=Gtk.FileChooserAction.SAVE,
            accept_label=_("_Export"),
            cancel_label=_("_Cancel")
        )
        self.export_dialog.set_current_name(_("Untitled plot") + ".png")
        self.export_dialog.connect("response", self.export_response)
        self.export_dialog.set_modal(True)
        self.export_dialog.show()

    def export_response(self, dialog, response):
        if response == Gtk.ResponseType.ACCEPT:
            self.gl_area.export_target = dialog.get_file().get_path()
            self.gl_area.queue_draw()
        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()
        self.export_dialog = None

    def delete_cb(self, window):
        self.prefs.save_config()

    def add_to_history(self, command):
        if self.can_redo():
            del self.history[self.history_position:]
        self.history.append(command)
        self.history_position = len(self.history)
        self.refresh_history_buttons()

    def can_undo(self):
        return self.history_position > 0

    def can_redo(self):
        return self.history_position < len(self.history)

    def undo(self, _):
        if self.can_undo():
            self.history_position -= 1
            self.history[self.history_position].undo(self)
            self.refresh_history_buttons()

    def redo(self, _):
        if self.history_position < len(self.history):
            self.history[self.history_position].do(self)
            self.history_position += 1
            self.refresh_history_buttons()

    def refresh_history_buttons(self):
        self.undo_button.props.sensitive = self.can_undo()
        self.redo_button.props.sensitive = self.can_redo()

    def prefs_updated(self, prefs):
        self.gl_area.queue_draw()


if __name__ == '__main__':
    Plots().run(sys.argv)



# Copyright 2021-2022 Alexander Huntley

# This file is part of Plots.

# Plots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Plots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Plots.  If not, see <https://www.gnu.org/licenses/>.

from jinja2 import Environment, PackageLoader

jinja_env = Environment(loader=PackageLoader('plots', 'shaders'))

GREEK_LETTERS = {
    'Alpha': 'Α',
    'Beta': 'Β',
    'Chi': 'Χ',
    'Delta': 'Δ',
    'Epsilon': 'Ε',
    'Eta': 'Η',
    'Gamma': 'Γ',
    'Iota': 'Ι',
    'Kappa': 'Κ',
    'Lambda': 'Λ',
    'Mu': 'Μ',
    'Nu': 'Ν',
    'Omega': 'Ω',
    'Omicron': 'Ο',
    'Phi': 'Φ',
    'Pi': 'Π',
    'Psi': 'Ψ',
    'Rho': 'Ρ',
    'Sigma': 'Σ',
    'Tau': 'Τ',
    'Theta': 'Θ',
    'Upsilon': 'Υ',
    'Xi': 'Ξ',
    'Zeta': 'Ζ',
    'alpha': 'α',
    'beta': 'β',
    'chi': 'χ',
    'delta': 'δ',
    'epsilon': 'ε',
    'eta': 'η',
    'gamma': 'γ',
    'iota': 'ι',
    'kappa': 'κ',
    'lambda': 'λ',
    'mu': 'μ',
    'nu': 'ν',
    'omega': 'ω',
    'omicron': 'ο',
    'phi': 'φ',
    "varphi": "φ",
    'pi': 'π',
    'psi': 'ψ',
    'rho': 'ρ',
    'sigma': 'σ',
    'tau': 'τ',
    'theta': 'θ',
    'upsilon': 'υ',
    'xi': 'ξ',
    'zeta': 'ζ'
}

GREEK_LETTERS_INVERSE = {char: name for name, char in GREEK_LETTERS.items()}
GREEK_REGEXES = GREEK_LETTERS.copy()
# stops a ψ being inserted while typing epsilon or upsilon
GREEK_REGEXES['(?<![EUeu])psi'] = GREEK_REGEXES.pop('psi')
# substrings must go second for regex to work
FUNCTIONS = "asech acsch acosech acoth sech csch cosech coth asec acsc acosec acot arcsec arccsc arccosec arccot sec csc cosec cot asinh acosh atanh sinh cosh tanh asin acos atan arcsin arccos arctan sinc sin cos tan exp log ln lg sign sgn".split()
BINARY_OPERATORS = ("+", "-", "*", "=")

SUP_ATOMS = "⁰¹²³⁴⁵⁶⁷⁸⁹ⁱⁿ"
SUP_CHARS = SUP_ATOMS + "⁺⁻⁽⁾"
SUP_TRAFO = str.maketrans(SUP_CHARS, "0123456789in+-()")
SUB_ATOMS = "₀₁₂₃₄₅₆₇₈₉ₐₑₒₓₕₖₗₘₙₚₛₜ"
SUB_CHARS = SUB_ATOMS + "₊₋₍₎"
SUB_TRAFO = str.maketrans(SUB_CHARS, "0123456789aeoxhklmnpst+-()")

